import os
from pathlib import Path

from .test_parsing import BaseLoaderTestCase

SCRIPT_DIRECTORY = os.path.abspath(os.path.dirname(__file__))
TEST_TOOL_DIRECTORIES = os.path.join(SCRIPT_DIRECTORY, "tool_directories")


TOOL_REQUIRED_FILES_XML_1 = """
<tool name="required_files_1" id="required_files_1" version="0.1.0">
    <command>foo</command>
    <required_files>
        <include path="my_script.R" />
    </required_files>
    <inputs>
        <input type="text" label="Text to parse." name="input1" />
    </inputs>
    <outputs>
        <output type="integer" name="out1" from="output" />
    </outputs>
</tool>
"""


TOOL_REQUIRED_FILES_XML_2 = """
<tool name="required_files_1" id="required_files_1" version="0.1.0">
    <command>foo</command>
    <required_files>
        <include path="*.R" type="glob" />
        <exclude path="other_script.R" />
    </required_files>
    <inputs>
        <input type="text" label="Text to parse." name="input1" />
    </inputs>
    <outputs>
        <output type="integer" name="out1" from="output" />
    </outputs>
</tool>
"""


TOOL_REQUIRED_FILES_XML_3 = """
<tool name="required_files_1" id="required_files_1" version="0.1.0">
    <command>foo</command>
    <required_files>
        <include path="*.R" type="glob" />
    </required_files>
    <inputs>
        <input type="text" label="Text to parse." name="input1" />
    </inputs>
    <outputs>
        <output type="integer" name="out1" from="output" />
    </outputs>
</tool>
"""


TOOL_REQUIRED_FILES_XML_4 = """
<tool name="required_files_1" id="required_files_1" version="0.1.0">
    <command>foo</command>
    <required_files>
        <include path=".*R" type="regex" />
        <exclude path="other_script*" type="glob" />
    </required_files>
    <inputs>
        <input type="text" label="Text to parse." name="input1" />
    </inputs>
    <outputs>
        <output type="integer" name="out1" from="output" />
    </outputs>
</tool>
"""


TOOL_REQUIRED_FILES_XML_DISABLED_DEFAULT_EXCLUSIONS = """
<tool name="required_files_1" id="required_files_1" version="0.1.0">
    <command>foo</command>
    <required_files extend_default_excludes="false">
        <include path="*.R" type="glob" />
    </required_files>
    <inputs>
        <input type="text" label="Text to parse." name="input1" />
    </inputs>
    <outputs>
        <output type="integer" name="out1" from="output" />
    </outputs>
</tool>
"""


class BaseRequiredFilesTestCase(BaseLoaderTestCase):
    source_file_name = "required_files.xml"

    def _required_files(self, tool_directory: str):
        tool_source = self._tool_source
        required_files = tool_source.parse_required_files()
        return required_files.find_required_files(tool_directory)


# directly include just one file that is there
class TestRequiredFiles1(BaseRequiredFilesTestCase):
    source_contents = TOOL_REQUIRED_FILES_XML_1

    def test_expected_files(self):
        files = self._required_files(os.path.join(TEST_TOOL_DIRECTORIES, "r-tool-dir"))
        assert len(files) == 1
        assert "my_script.R" in files


# include a glob and exclude a file in the glob
class TestRequiredFiles2(BaseRequiredFilesTestCase):
    source_contents = TOOL_REQUIRED_FILES_XML_2

    def test_expected_files(self):
        files = self._required_files(os.path.join(TEST_TOOL_DIRECTORIES, "r-tool-dir"))
        assert len(files) == 1
        assert "my_script.R" in files


# include a glob with multiple matches
class TestRequiredFiles3(BaseRequiredFilesTestCase):
    source_contents = TOOL_REQUIRED_FILES_XML_3

    def test_expected_files(self):
        files = self._required_files(os.path.join(TEST_TOOL_DIRECTORIES, "r-tool-dir"))
        assert len(files) == 2
        assert "my_script.R" in files
        assert "other_script.R" in files


# include a file with regex and exclude with glob
class TestRequiredFiles4(BaseRequiredFilesTestCase):
    source_contents = TOOL_REQUIRED_FILES_XML_4

    def test_expected_files(self):
        files = self._required_files(os.path.join(TEST_TOOL_DIRECTORIES, "r-tool-dir"))
        assert len(files) == 1
        assert "my_script.R" in files


class TestHgExcludedByDefault(BaseRequiredFilesTestCase):
    source_contents = TOOL_REQUIRED_FILES_XML_3

    def test_expected_files(self):
        repo_dir = setup_dir_with_repo(self.temp_directory)
        files = self._required_files(repo_dir)
        assert len(files) == 1
        assert "my_script.R" in files


class TestHgExclusionDisabled(BaseRequiredFilesTestCase):
    source_contents = TOOL_REQUIRED_FILES_XML_DISABLED_DEFAULT_EXCLUSIONS

    def test_expected_files(self):
        repo_dir = setup_dir_with_repo(self.temp_directory)
        files = self._required_files(repo_dir)
        assert len(files) == 2
        assert "my_script.R" in files
        assert ".hg/index.R" in files


def setup_dir_with_repo(tmp_dir):
    repo = os.path.join(tmp_dir, "repo")
    os.makedirs(repo)
    hg = os.path.join(repo, ".hg")
    os.makedirs(hg)
    Path(hg, "index.R").touch()
    Path(repo, "my_script.R").touch()
    return repo



import tempfile

import pytest
import yaml

from galaxy.tool_util.deps.mulled.mulled_build import target_str_to_targets
from galaxy.tool_util.deps.mulled.mulled_build_files import (
    generate_targets,
    Target,
)

TESTCASES = yaml.safe_load(r"""
- test_legacy_files_package_only:
    content: samtools
    equals:
      base_image: null
      image_build: null
      name_override: null
      targets: samtools
- test_legacy_files_package_image_build:
    content: "samtools\t10"
    equals:
      base_image: null
      image_build: '10'
      name_override: null
      targets: samtools
- test_legacy_files_package_image_build_name_override:
    content: "samtools\t10\timage_name"
    equals:
      base_image: null
      image_build: '10'
      name_override: image_name
      targets: samtools
- test_files_package_image_build_name_override_base_image_with_header:
    content: "#targets\timage_build\tname_override\tbase_image\nsamtools\t10\timage_name\textended_image"
    equals:
      base_image: extended_image
      image_build: '10'
      name_override: image_name
      targets: samtools
- test_files_package_image_build_base_image_with_header:
    content: "#targets\timage_build\tbase_image\nsamtools\t10\textended_image"
    equals:
      base_image: extended_image
      image_build: '10'
      name_override: null
      targets: samtools
- test_files_package_image_build_name_override_base_image_with_header_reordered:
    content: "#base_image\ttargets\timage_build\tname_override\nextended_image\tsamtools\t10\timage_name"
    equals:
      base_image: extended_image
      image_build: '10'
      name_override: image_name
      targets: samtools
""")
TEST_IDS = [next(iter(k.keys())) for k in TESTCASES]


@pytest.mark.parametrize(
    "content, equals", [(d[k]["content"], d[k]["equals"]) for k, d in zip(TEST_IDS, TESTCASES)], ids=TEST_IDS
)
def test_generate_targets(content, equals):
    equals["targets"] = target_str_to_targets(equals["targets"])
    equals = Target(**equals)
    with tempfile.NamedTemporaryFile(mode="w") as tmpfile:
        tmpfile.write(content)
        tmpfile.flush()
        generated_target = next(generate_targets(tmpfile.name))
    assert generated_target.targets == equals.targets
    assert generated_target.image_build == equals.image_build
    assert generated_target.name_override == equals.name_override
    assert generated_target.base_image == equals.base_image



import os
import re
from shutil import rmtree
from tempfile import mkdtemp

from galaxy.tool_util.loader import (
    load_tool,
    template_macro_params,
)
from galaxy.tool_util.unittest_utils.sample_data import (
    SIMPLE_MACRO,
    SIMPLE_TOOL_WITH_MACRO,
)
from galaxy.util import (
    parse_xml,
    xml_to_string,
)


class TestToolDirectory:
    __test__ = False  # Prevent pytest from discovering this class (issue #12071)

    def __init__(self):
        self.temp_directory = mkdtemp()

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        rmtree(self.temp_directory)

    def write(self, contents, name="tool.xml"):
        open(os.path.join(self.temp_directory, name), "w").write(contents)

    def load(self, name="tool.xml", preprocess=True):
        path = os.path.join(self.temp_directory, name)
        if preprocess:
            return load_tool(path)
        else:
            return parse_xml(path)


def test_no_macros():
    """
    Test tool loaded in absence of a macros node.
    """
    with TestToolDirectory() as tool_dir:
        tool_dir.write("<tool/>")
        tool_dir.load(preprocess=True)


def test_loader_simple():
    """
    Test simple macro replacement.
    """
    with TestToolDirectory() as tool_dir:
        tool_dir.write("""
<tool>
    <expand macro="inputs" />
    <macros>
        <macro name="inputs">
            <inputs />
        </macro>
    </macros>
</tool>""")
        xml = tool_dir.load(preprocess=False)
        assert xml.find("inputs") is None
        xml = tool_dir.load(preprocess=True)
        assert xml.find("inputs") is not None


def test_loader_external():
    """
    Test importing macros from external files
    """
    with TestToolDirectory() as tool_dir:
        tool_dir.write(SIMPLE_TOOL_WITH_MACRO)

        tool_dir.write(SIMPLE_MACRO.substitute(tool_version="2.0"), name="external.xml")
        xml = tool_dir.load(preprocess=False)
        assert xml.find("inputs") is None
        xml = tool_dir.load(preprocess=True)
        assert xml.find("inputs") is not None


def test_loader_unnamed_yield():
    # Test macros with unnamed yield statements.
    with TestToolDirectory() as tool_dir:
        tool_dir.write("""
<tool>
    <expand macro="inputs">
        <input name="first_input" />
    </expand>
    <expand macro="inputs">
        <input name="second_input" />
    </expand>
    <expand macro="inputs">
        <input name="third_input" />
    </expand>
    <macros>
        <macro name="inputs">
            <expand macro="foo">
                <yield />
            </expand>
        </macro>
        <macro name="foo">
            <inputs>
                <yield />
            </inputs>
        </macro>
    </macros>
</tool>""")
        xml = tool_dir.load()
        assert xml.find("/inputs[1]/input").get("name") == "first_input"
        assert xml.find("/inputs[2]/input").get("name") == "second_input"
        assert xml.find("/inputs[3]/input").get("name") == "third_input"


def test_loader_unnamed_yield_nested():
    """
    Test nested macro with yield statements
    """
    with TestToolDirectory() as tool_dir:
        tool_dir.write("""
<tool>
    <macros>
        <macro name="paired_options">
            <when value="paired">
                <yield />
            </when>
            <when value="paired_collection">
                <yield />
            </when>
        </macro>
        <macro name="single_or_paired_general">
            <conditional name="library">
                <expand macro="paired_options">
                    <yield />
                </expand>
            </conditional>
        </macro>
    </macros>
    <inputs>
        <expand macro="single_or_paired_general">
            <param name="test"/>
        </expand>
    </inputs>
</tool>
""")
        xml = tool_dir.load()
        # assert the both yields in the inner macro (paired_options) are expanded
        assert xml.find('/inputs/conditional[@name="library"]/when[@value="paired"]/param[@name="test"]') is not None
        assert (
            xml.find('/inputs/conditional[@name="library"]/when[@value="paired_collection"]/param[@name="test"]')
            is not None
        )


def test_loader_recursive():
    """
    Test recursive macro applications.
    """
    with TestToolDirectory() as tool_dir:
        tool_dir.write("""
<tool>
    <expand macro="inputs">
        <input name="first_input" />
        <expand macro="second" />
    </expand>
    <macros>
        <macro name="inputs">
            <inputs>
                <yield />
                <input name="third_input" />
            </inputs>
        </macro>
        <macro name="second">
            <input name="second_input" />
        </macro>
    </macros>
</tool>""")
        xml = tool_dir.load()
        assert xml.find("/inputs/input[1]").get("name") == "first_input"
        assert xml.find("/inputs/input[2]").get("name") == "second_input"
        assert xml.find("/inputs/input[3]").get("name") == "third_input"


def test_loader_recursive2():
    """
    Test recursive macro applications.
    """
    with TestToolDirectory() as tool_dir:
        tool_dir.write("""
<tool>
    <expand macro="inputs">
        <input name="first_input" />
        <expand macro="second" />
    </expand>
    <macros>
        <macro name="inputs">
            <inputs>
                <yield />
            </inputs>
        </macro>
        <macro name="second">
            <expand macro="second_delegate" />
            <input name="third_input" />
        </macro>
        <macro name="second_delegate">
            <input name="second_input" />
        </macro>
    </macros>
</tool>""")
        xml = tool_dir.load()
        assert xml.find("/inputs/input[1]").get("name") == "first_input"
        assert xml.find("/inputs/input[2]").get("name") == "second_input"
        assert xml.find("/inputs/input[3]").get("name") == "third_input"


def test_loader_toplevel_yield():
    """
    test expansion of top level (ie child of <macro>) yields
    """
    with TestToolDirectory() as tool_dir:
        tool_dir.write("""
<tool id="issue_647">
    <macros>
        <macro name="a">
            <param name="a1" type="text" value="a1" label="a1"/>
            <yield />
        </macro>
    </macros>
    <inputs>
        <expand macro="a">
            <param name="b" type="text" value="b" label="b" />
        </expand>
    </inputs>
</tool>""")
        xml = tool_dir.load()
        assert xml.find("/inputs/param[1]").get("name") == "a1"
        assert xml.find("/inputs/param[2]").get("name") == "b"


def test_loader_shortcut():
    """
    Test <xml> is shortcut for macro type="xml"
    """
    with TestToolDirectory() as tool_dir:
        tool_dir.write("""
<tool>
    <expand macro="inputs" />
    <macros>
        <xml name="inputs">
            <inputs />
        </xml>
    </macros>
</tool>""")
        xml = tool_dir.load()
        assert xml.find("inputs") is not None


def test_loader_template():
    with TestToolDirectory() as tool_dir:
        tool_dir.write("""
<tool>
    <command interpreter="python">tool_wrapper.py
    #include source=$tool_params
    </command>
    <macros>
        <template name="tool_params">-a 1 -b 2</template>
    </macros>
</tool>
""")
        xml = tool_dir.load()
        params_dict = template_macro_params(xml.getroot())
        assert params_dict["tool_params"] == "-a 1 -b 2"


def test_loader_token_text():
    with TestToolDirectory() as tool_dir:
        tool_dir.write("""
<tool>
    <macros>
        <token name="@CITATION@">The citation.</token>
    </macros>
    <help>@CITATION@</help>
    <another>
        <tag />
    </another>
</tool>
""")
        xml = tool_dir.load()
        help_el = xml.find("help")
        assert help_el.text == "The citation.", help_el.text


def test_loader_token_nested():
    with TestToolDirectory() as tool_dir:
        tool_dir.write("""
<tool>
    <macros>
        <token name="@WIZARD@">Ab@HELPER@ra</token>
        <token name="@HELPER@">ra@WITCH@ab</token>
        <token name="@WITCH@">cad</token>
    </macros>
    <help>@WIZARD@</help>
    <another>
        <tag />
    </another>
</tool>
""")
        xml = tool_dir.load()
        help_el = xml.find("help")
        assert help_el.text == "Abracadabra", help_el.text


def test_loader_token_cycle():
    """
    test if cycles in nested tokens are detected
    """
    with TestToolDirectory() as tool_dir:
        tool_dir.write("""
<tool>
    <macros>
        <token name="@A@">a @ROSE@</token>
        <token name="@ROSE@">rose @IS@</token>
        <token name="@IS@">a @A@</token>
    </macros>
    <help>@A@</help>
    <another>
        <tag />
    </another>
</tool>
""")
        try:
            tool_dir.load()
        except Exception as e:
            assert str(e) == "Token '@IS@' cannot contain itself"
        else:
            raise AssertionError("Cycle not detected, but then we are doomed anyway")


def test_loader_token_attribute_name():
    """
    test the repacement of an attribute name by a token
    """
    with TestToolDirectory() as tool_dir:
        tool_dir.write("""
<tool>
    <macros>
        <token name="__ATTR_NAME">name</token>
    </macros>
    <another>
        <tag __ATTR_NAME="blah" />
    </another>
</tool>
""")
        xml = tool_dir.load()
        assert xml.find('/another/tag[@name="blah"]') is not None


def test_loader_token_attribute_value():
    """
    test the repacement of an attribute value by a token
    """
    with TestToolDirectory() as tool_dir:
        tool_dir.write("""
<tool>
    <macros>
        <token name="@TAG_VAL@">The value.</token>
    </macros>
    <another>
        <tag value="@TAG_VAL@" />
    </another>
</tool>
""")
        xml = tool_dir.load()
        assert xml.find('/another/tag[@value="The value."]') is not None


def test_loader_token_empty():
    with TestToolDirectory() as tool_dir:
        tool_dir.write("""
<tool>
    <macros>
        <token name="@TAG_VAL@"><![CDATA[]]></token>
    </macros>
    <another>
        <tag value="@TAG_VAL@" />
    </another>
</tool>
""")
        xml = tool_dir.load()
        assert xml.find('/another/tag[@value=""]') is not None


def test_loader_macro_token_quote():
    """
    Test macros XML macros with $$ expansions in attributes
    """
    with TestToolDirectory() as tool_dir:
        tool_dir.write("""
<tool>
    <expand macro="inputs" bar="hello" />
    <expand macro="inputs" bar="my awesome" />
    <expand macro="inputs" bar="doggo" />
    <macros>
        <xml name="inputs" tokens="bar" token_quote="$$">
            <inputs type="the type is $$BAR$$" />
        </xml>
    </macros>
</tool>
""")
        xml = tool_dir.load()
        input_els = xml.findall("inputs")
        assert len(input_els) == 3
        assert input_els[0].attrib["type"] == "the type is hello"
        assert input_els[1].attrib["type"] == "the type is my awesome"
        assert input_els[2].attrib["type"] == "the type is doggo"


def test_loader_macro_token():
    """
    Test macros XML macros with @ expansions in text
    """
    with TestToolDirectory() as tool_dir:
        tool_dir.write("""
<tool>
    <expand macro="inputs" foo="hello" />
    <expand macro="inputs" foo="world" />
    <expand macro="inputs" />
    <macros>
        <xml name="inputs" token_foo="the_default">
            <inputs>@FOO@</inputs>
        </xml>
    </macros>
</tool>
""")
        xml = tool_dir.load()
        input_els = xml.findall("inputs")
        assert len(input_els) == 3
        assert input_els[0].text == "hello"
        assert input_els[1].text == "world"
        assert input_els[2].text == "the_default"


def test_loader_macro_token_recursive():
    """
    Test macros XML macros with @ expansions and recursive
    """
    with TestToolDirectory() as tool_dir:
        tool_dir.write("""
<tool>
    <expand macro="inputs" foo="hello" />
    <expand macro="inputs" foo="world" />
    <expand macro="inputs" />
    <macros>
        <xml name="inputs" token_foo="the_default">
            <expand macro="real_inputs"><cow>@FOO@</cow></expand>
        </xml>
        <xml name="real_inputs">
            <inputs><yield /></inputs>
        </xml>
    </macros>
</tool>
""")
        xml = tool_dir.load()
        input_els = xml.findall("inputs")
        assert len(input_els) == 3
        assert input_els[0].find("cow").text == "hello"
        assert input_els[1].find("cow").text == "world"
        assert input_els[2].find("cow").text == "the_default"


def test_loader_macro_named_yield():
    """
    test expansion of named and unnamed yield
    - named yields are replaced by content of the corresponding token
    - unnamed yields are replaced by all non-token elements of the expand tag
    """
    with TestToolDirectory() as tool_dir:
        tool_dir.write("""
<tool>
    <macros>
        <xml name="test">
            <A>
                <yield/>
            </A>
            <yield name="token1"/>
            <B>
                <yield/>
                <yield name="token2"/>
            </B>
        </xml>
    </macros>
    <expand macro="test">
        <token name="token1">
            <content_of_token1/>
            <more_content_of_token1/>
        </token>
        <sub_of_expand_1/>
        <token name="token2">
            <content_of_token2/>
            <more_content_of_token2/>
        </token>
        <sub_of_expand_2/>
    </expand>
</tool>
""")
        xml = tool_dir.load()
        assert xml_to_string(xml, pretty=True) == """<?xml version="1.0" ?>
<tool>
    <macros/>
    <A>
        <sub_of_expand_1/>
        <sub_of_expand_2/>
    </A>
    <content_of_token1/>
    <more_content_of_token1/>
    <B>
        <sub_of_expand_1/>
        <sub_of_expand_2/>
        <content_of_token2/>
        <more_content_of_token2/>
    </B>
</tool>"""


def test_loader_macro_multiple_toplevel_yield():
    """
    test replacement of multiple top level yield
    """
    with TestToolDirectory() as tool_dir:
        tool_dir.write("""
<tool>
    <macros>
        <xml name="test">
            <blah/>
            <yield/>
            <blah>
                <yield name="token1"/>
            </blah>
            <yield name="token2"/>
        </xml>
    </macros>
    <expand macro="test">
        <token name="token1">
            <content_of_token1/>
            <more_content_of_token1/>
        </token>
        <sub_of_expand_1/>
        <token name="token2">
            <content_of_token2/>
            <more_content_of_token2/>
        </token>
        <sub_of_expand_2/>
    </expand>
</tool>
""")
        xml = tool_dir.load()
        assert xml_to_string(xml, pretty=True) == """<?xml version="1.0" ?>
<tool>
    <macros/>
    <blah/>
    <sub_of_expand_1/>
    <sub_of_expand_2/>
    <blah>
        <content_of_token1/>
        <more_content_of_token1/>
    </blah>
    <content_of_token2/>
    <more_content_of_token2/>
</tool>"""


def test_loader_macro_recursive_named_yield():
    """
    test 'recursive' replacement with named yields
    since named yields are processed in order of the definition of the
    corresponding tokens:
    - replacing yield for token1 introduces yield for token2
    - replacing yield for token2 introduced unnamed yield
    - replacing unnamed yield gives the only non-token element of the expand
    """
    with TestToolDirectory() as tool_dir:
        tool_dir.write("""
<tool>
    <macros>
        <xml name="test">
            <A>
                <yield name="token1"/>
            </A>
        </xml>
    </macros>
    <expand macro="test">
        <token name="token1">
            <T1>
                <yield name="token2"/>
            </T1>
        </token>
        <token name="token2">
            <T2>
                <yield/>
            </T2>
        </token>
        <T/>
    </expand>
</tool>""")

        xml = tool_dir.load()
        assert xml_to_string(xml, pretty=True) == """<?xml version="1.0" ?>
<tool>
    <macros/>
    <A>
        <T1>
            <T2>
                <T/>
            </T2>
        </T1>
    </A>
</tool>"""


def test_loader_specify_nested_macro_by_token():
    """
    test if a nested macro can have a nested
    macro specifying the macro name via a token
    of the outer macro
    """
    with TestToolDirectory() as tool_dir:
        tool_dir.write("""
<tool>
    <macros>
        <import>external.xml</import>
    </macros>
    <expand macro="testp" token_name="a"/>
    <expand macro="testp" token_name="b"/>
</tool>""")
        tool_dir.write(
            """
<macros>
    <xml name="a">
        <A/>
    </xml>
    <xml name="b">
        <B/>
    </xml>
    <xml name="testp" tokens="token_name">
        <expand macro="@TOKEN_NAME@"/>
    </xml>
</macros>
""",
            name="external.xml",
        )

        xml = tool_dir.load()
        # test with re because loading from external macros
        # adds a xml:base property (containing the source path)
        # to the node which is printed
        print(f"{xml_to_string(xml, pretty=True)}")
        assert re.match(
            r"""<\?xml version="1\.0" \?>
<tool>
    <macros/>
    <A/>
    <B/>
</tool>""",
            xml_to_string(xml, pretty=True),
            re.MULTILINE,
        )


def test_loader_circular_macros():
    """
    check the cycles in nested macros are detected
    """
    with TestToolDirectory() as tool_dir:
        tool_dir.write("""
<tool>
    <macros>
        <xml name="a">
            <A>
                <expand macro="b"/>
            </A>
        </xml>
        <xml name="b">
            <B>
                <expand macro="a"/>
            </B>
        </xml>
    </macros>
    <expand macro="a"/>
</tool>""")
        try:
            tool_dir.load()
        except AssertionError as a:
            assert str(a) == "Cycle in nested macros: already expanded ['a', 'b'] can't expand 'a' again"



import os.path
import tempfile
from contextlib import contextmanager
from os import (
    chmod,
    makedirs,
    stat,
    symlink,
)
from shutil import rmtree
from stat import S_IXUSR
from subprocess import (
    PIPE,
    Popen,
)
from typing import (
    Any,
    Dict,
    Optional,
)

from galaxy.tool_util.deps import (
    build_dependency_manager,
    DependencyManager,
)
from galaxy.tool_util.deps.requirements import (
    ToolRequirement,
    ToolRequirements,
)
from galaxy.tool_util.deps.resolvers import NullDependency
from galaxy.tool_util.deps.resolvers.galaxy_packages import GalaxyPackageDependency
from galaxy.tool_util.deps.resolvers.lmod import (
    LmodDependency,
    LmodDependencyResolver,
)
from galaxy.tool_util.deps.resolvers.modules import (
    ModuleDependency,
    ModuleDependencyResolver,
)
from galaxy.util import unicodify
from galaxy.util.bunch import Bunch
from .util import modify_environ

# If true, test created DependencyManager objects by serializing out to json and re-constituting.
ROUND_TRIP_TEST_DEPENDENCY_MANAGER_SERIALIZATION = True


def test_tool_dependencies():
    # Setup directories

    with __test_base_path() as base_path:
        for name, version, sub in [("dep1", "1.0", "env.sh"), ("dep1", "2.0", "bin"), ("dep2", "1.0", None)]:
            if sub == "bin":
                p = os.path.join(base_path, name, version, "bin")
            else:
                p = os.path.join(base_path, name, version)
            try:
                makedirs(p)
            except Exception:
                pass
            if sub == "env.sh":
                __touch(os.path.join(p, "env.sh"))

        dm = __dependency_manager_for_base_path(default_base_path=base_path)
        dependency = dm.find_dep("dep1", "1.0")
        assert dependency.script == os.path.join(base_path, "dep1", "1.0", "env.sh")
        assert dependency.path == os.path.join(base_path, "dep1", "1.0")
        assert dependency.version == "1.0"
        dependency = dm.find_dep("dep1", "2.0")
        assert dependency.script is None
        assert dependency.path == os.path.join(base_path, "dep1", "2.0")
        assert dependency.version == "2.0"

        # Test default versions
        symlink(os.path.join(base_path, "dep1", "2.0"), os.path.join(base_path, "dep1", "default"))
        dependency = dm.find_dep("dep1", None)
        assert dependency.version == "2.0"

        # Test default resolve will be fall back on default package dependency
        # when using the default resolver.
        dependency = dm.find_dep("dep1", "2.1")
        assert dependency.version == "2.0"  # 2.0 is defined as default_version


TEST_REPO_USER = "devteam"
TEST_REPO_NAME = "bwa"
TEST_REPO_CHANGESET = "12abcd41223da"
TEST_VERSION = "0.5.9"


def test_toolshed_set_enviornment_requiremetns():
    with __test_base_path() as base_path:
        test_repo = __build_test_repo("set_environment")
        dm = __dependency_manager_for_base_path(default_base_path=base_path)
        env_settings_dir = os.path.join(
            base_path, "environment_settings", TEST_REPO_NAME, TEST_REPO_USER, TEST_REPO_NAME, TEST_REPO_CHANGESET
        )
        os.makedirs(env_settings_dir)
        dependency = dm.find_dep(
            TEST_REPO_NAME, version=None, type="set_environment", installed_tool_dependencies=[test_repo]
        )
        assert dependency.version is None
        assert dependency.script == os.path.join(env_settings_dir, "env.sh")


def test_toolshed_package_requirements():
    with __test_base_path() as base_path:
        test_repo = __build_test_repo("package", version=TEST_VERSION)
        dm = __dependency_manager_for_base_path(default_base_path=base_path)
        package_dir = __build_ts_test_package(base_path)
        dependency = dm.find_dep(
            TEST_REPO_NAME, version=TEST_VERSION, type="package", installed_tool_dependencies=[test_repo]
        )
        assert dependency.version == TEST_VERSION
        assert dependency.script == os.path.join(package_dir, "env.sh")


def test_toolshed_tools_fallback_on_manual_dependencies():
    with __test_base_path() as base_path:
        dm = __dependency_manager_for_base_path(default_base_path=base_path)
        test_repo = __build_test_repo("package", version=TEST_VERSION)
        env_path = __setup_galaxy_package_dep(base_path, "dep1", "1.0")
        dependency = dm.find_dep("dep1", version="1.0", type="package", installed_tool_dependencies=[test_repo])
        assert dependency.version == "1.0"
        assert dependency.script == env_path


def test_toolshed_greater_precendence():
    with __test_base_path() as base_path:
        dm = __dependency_manager_for_base_path(default_base_path=base_path)
        test_repo = __build_test_repo("package", version=TEST_VERSION)
        ts_package_dir = __build_ts_test_package(base_path)
        gx_env_path = __setup_galaxy_package_dep(base_path, TEST_REPO_NAME, TEST_VERSION)
        ts_env_path = os.path.join(ts_package_dir, "env.sh")
        dependency = dm.find_dep(
            TEST_REPO_NAME, version=TEST_VERSION, type="package", installed_tool_dependencies=[test_repo]
        )
        assert dependency.script != gx_env_path  # Not the galaxy path, it should be the tool shed path used.
        assert dependency.script == ts_env_path


def __build_ts_test_package(base_path, script_contents=""):
    package_dir = os.path.join(
        base_path, TEST_REPO_NAME, TEST_VERSION, TEST_REPO_USER, TEST_REPO_NAME, TEST_REPO_CHANGESET
    )
    __touch(os.path.join(package_dir, "env.sh"), script_contents)
    return package_dir


REQUIREMENT_A = {"name": "gnuplot", "type": "package", "version": "4.6"}
REQUIREMENT_B = REQUIREMENT_A.copy()
REQUIREMENT_B["version"] = "4.7"


def test_tool_requirement_equality():
    a = ToolRequirement.from_dict(REQUIREMENT_A)
    assert a == ToolRequirement(**REQUIREMENT_A)  # type: ignore[arg-type] # https://github.com/python/mypy/issues/10008
    b = ToolRequirement(**REQUIREMENT_B)  # type: ignore[arg-type] # https://github.com/python/mypy/issues/10008
    assert a != b


def test_tool_requirements():
    tool_requirements_ab = ToolRequirements([REQUIREMENT_A, REQUIREMENT_B])
    tool_requirements_ab_dup = ToolRequirements([REQUIREMENT_A, REQUIREMENT_B])
    tool_requirements_b = ToolRequirements([REQUIREMENT_A])
    assert tool_requirements_ab == ToolRequirements([REQUIREMENT_B, REQUIREMENT_A])
    assert tool_requirements_ab == ToolRequirements([REQUIREMENT_B, REQUIREMENT_A, REQUIREMENT_A])
    assert tool_requirements_ab != tool_requirements_b
    assert len({tool_requirements_ab, tool_requirements_ab_dup}) == 1


def test_module_dependency_resolver():
    with __test_base_path() as temp_directory:
        module_script = _setup_module_command(
            temp_directory,
            """
-------------------------- /soft/modules/modulefiles ---------------------------
JAGS/3.2.0-gcc45
JAGS/3.3.0-gcc4.7.2
ProbABEL/0.1-3
ProbABEL/0.1-9e
R/2.12.2
R/2.13.1
R/2.14.1
R/2.15.0
R/2.15.1
R/3.0.1(default)
abokia-blast/2.0.2-130524/ompi_intel
abokia-blast/2.0.2-130630/ompi_intel

--------------------------- /soft/intel/modulefiles ----------------------------
advisor/2013/update1    intel/11.1.075          mkl/10.2.1.017
advisor/2013/update2    intel/11.1.080          mkl/10.2.5.035
advisor/2013/update3    intel/12.0              mkl/10.2.7.041
""",
        )
        resolver = ModuleDependencyResolver(_SimpleDependencyManager(), modulecmd=module_script)
        module = resolver.resolve(ToolRequirement(name="R", version=None, type="package"))
        assert module.module_name == "R"
        assert module.module_version is None

        module = resolver.resolve(ToolRequirement(name="R", version="3.0.1", type="package"))
        assert module.module_name == "R"
        assert module.module_version == "3.0.1"

        module = resolver.resolve(ToolRequirement(name="R", version="3.0.4", type="package"))
        assert isinstance(module, NullDependency)


def test_module_resolver_with_mapping():
    with __test_base_path() as temp_directory:
        module_script = _setup_module_command(
            temp_directory,
            """
-------------------------- /soft/modules/modulefiles ---------------------------
blast/2.24
""",
        )
        mapping_file = os.path.join(temp_directory, "mapping")
        with open(mapping_file, "w") as f:
            f.write("""
- from: blast+
  to: blast
""")

        resolver = ModuleDependencyResolver(
            _SimpleDependencyManager(), modulecmd=module_script, mapping_files=mapping_file
        )
        module = resolver.resolve(ToolRequirement(name="blast+", version="2.24", type="package"))
        assert module.module_name == "blast"
        assert module.module_version == "2.24", module.module_version


def test_module_resolver_with_mapping_specificity_rules():
    # If a requirement demands a specific version,
    # do not map to a different version when the version
    # has not been specified explicitly
    with __test_base_path() as temp_directory:
        module_script = _setup_module_command(
            temp_directory,
            """
-------------------------- /soft/modules/modulefiles ---------------------------
blast/2.24
""",
        )
        mapping_file = os.path.join(temp_directory, "mapping")
        with open(mapping_file, "w") as f:
            f.write("""
- from:
    name: blast
    unversioned: true
  to:
    name: blast
    version: 2.24
""")

        resolver = ModuleDependencyResolver(
            _SimpleDependencyManager(), modulecmd=module_script, mapping_files=mapping_file
        )
        module = resolver.resolve(ToolRequirement(name="blast", type="package"))
        assert module.module_name == "blast"
        assert (
            module.module_version == "2.24"
        ), module.module_version  # successful resolution, because Requirement does not ask for a specific version
        module = resolver.resolve(ToolRequirement(name="blast", version="2.22", type="package"))
        assert isinstance(
            module, NullDependency
        )  # NullDependency, because we don't map `version: Null` over a Requirement that asks for a specific version


def test_module_resolver_with_mapping_versions():
    with __test_base_path() as temp_directory:
        module_script = _setup_module_command(
            temp_directory,
            """
-------------------------- /soft/modules/modulefiles ---------------------------
blast/2.22.0-mpi
blast/2.23
blast/2.24.0-mpi
""",
        )
        mapping_file = os.path.join(temp_directory, "mapping")
        with open(mapping_file, "w") as f:
            f.write("""
- from:
    name: blast+
    version: 2.24
  to:
    name: blast
    version: 2.24.0-mpi
- from:
    name: blast
    version: 2.22
  to:
    version: 2.22.0-mpi
""")

        resolver = ModuleDependencyResolver(
            _SimpleDependencyManager(), modulecmd=module_script, mapping_files=mapping_file
        )
        module = resolver.resolve(ToolRequirement(name="blast+", version="2.24", type="package"))
        assert module.module_name == "blast"
        assert module.module_version == "2.24.0-mpi", module.module_version

        resolver = ModuleDependencyResolver(
            _SimpleDependencyManager(), modulecmd=module_script, mapping_files=mapping_file
        )
        module = resolver.resolve(ToolRequirement(name="blast+", version="2.23", type="package"))
        assert isinstance(module, NullDependency)

        module = resolver.resolve(ToolRequirement(name="blast", version="2.22", type="package"))
        assert module.module_name == "blast"
        assert module.module_version == "2.22.0-mpi", module.module_version


def test_module_dependency_resolver_skip_availability_check():
    with __test_base_path() as temp_directory:
        module_script = _setup_module_command(
            temp_directory,
            """
-------------------------- /soft/modules/modulefiles ---------------------------
BlastPlus/2.4.0+
Infernal/1.1.2
Mothur/1.36.1
""",
        )
        resolver = ModuleDependencyResolver(
            _SimpleDependencyManager(), modulecmd=module_script, skip_availability_check="true"
        )

        module = resolver.resolve(ToolRequirement(name="Mothur", version="1.36.1", type="package"))
        assert module.module_name == "Mothur"
        assert module.module_version == "1.36.1"

        module = resolver.resolve(ToolRequirement(name="BlastPlus", version="2.3", type="package"))
        assert module.module_name == "BlastPlus"
        assert module.module_version == "2.3"

        module = resolver.resolve(ToolRequirement(name="Infernal", version=None, type="package"))
        assert module.module_name == "Infernal"
        assert module.module_version is None

        module = resolver.resolve(ToolRequirement(name="Foo", version="0.1", type="package"))
        assert module.module_name == "Foo"
        assert module.module_version == "0.1"

        module = resolver.resolve(ToolRequirement(name="Bar", version=None, type="package"))
        assert module.module_name == "Bar"
        assert module.module_version is None


def _setup_module_command(temp_directory, contents):
    module_script = os.path.join(temp_directory, "modulecmd")
    __write_script(
        module_script,
        f"""#!/bin/sh
cat {temp_directory}/example_output 1>&2;
""",
    )
    with open(os.path.join(temp_directory, "example_output"), "w") as f:
        # Subset of module avail from MSI cluster.
        f.write(contents)
    return module_script


def test_module_dependency():
    with __test_base_path() as temp_directory:
        # Create mock modulecmd script that just exports a variable
        # the way modulecmd sh load would, but also validate correct
        # module name and version are coming through.
        mock_modulecmd = os.path.join(temp_directory, "modulecmd")
        __write_script(
            mock_modulecmd,
            """#!/bin/sh
if [ $3 != "foomodule/1.0" ];
then
    exit 1
fi
echo 'FOO="bar"'
""",
        )
        resolver = Bunch(modulecmd=mock_modulecmd, modulepath="/something")
        dependency = ModuleDependency(resolver, "foomodule", "1.0")
        __assert_foo_exported(dependency.shell_commands())


def test_lmod_dependency_resolver():
    with __test_base_path() as temp_directory:
        lmod_script = _setup_lmod_command(
            temp_directory,
            """
/opt/apps/modulefiles:
BlastPlus/
BlastPlus/2.2.31+
BlastPlus/2.4.0+
Infernal/
Infernal/1.1.2
Mothur/
Mothur/1.33.3
Mothur/1.36.1
Mothur/1.38.1.1
""",
        )
        resolver = LmodDependencyResolver(
            _SimpleDependencyManager(), lmodexec=lmod_script, modulepath="/path/to/modulefiles"
        )

        lmod = resolver.resolve(ToolRequirement(name="Infernal", version=None, type="package"))
        assert lmod.module_name == "Infernal"
        assert lmod.module_version is None

        lmod = resolver.resolve(ToolRequirement(name="BlastPlus", version="2.4.0+", type="package"))
        assert lmod.module_name == "BlastPlus"
        assert lmod.module_version == "2.4.0+"

        lmod = resolver.resolve(ToolRequirement(name="Mothur", version="1.39", type="package"))
        assert isinstance(lmod, NullDependency)


def test_lmod_dependency_resolver_versionless():
    with __test_base_path() as temp_directory:
        lmod_script = _setup_lmod_command(
            temp_directory,
            """
/opt/apps/modulefiles:
BlastPlus/2.4.0+
Infernal/1.1.2
Mothur/1.36.1
""",
        )
        resolver = LmodDependencyResolver(
            _SimpleDependencyManager(), lmodexec=lmod_script, versionless="true", modulepath="/path/to/modulefiles"
        )

        lmod = resolver.resolve(ToolRequirement(name="Infernal", version=None, type="package"))
        assert lmod.module_name == "Infernal"
        assert lmod.module_version is None

        lmod = resolver.resolve(ToolRequirement(name="Mothur", version="1.36.1", type="package"))
        assert lmod.module_name == "Mothur"
        assert lmod.module_version == "1.36.1"

        lmod = resolver.resolve(ToolRequirement(name="BlastPlus", version="2.3", type="package"))
        assert lmod.module_name == "BlastPlus"
        assert lmod.module_version is None

        lmod = resolver.resolve(ToolRequirement(name="Foo", version="0.1", type="package"))
        assert isinstance(lmod, NullDependency)


def test_lmod_dependency_resolver_with_mapping_file():
    with __test_base_path() as temp_directory:
        lmod_script = _setup_lmod_command(
            temp_directory,
            """
/opt/apps/modulefiles:
BlastPlus/
BlastPlus/2.2.31+
BlastPlus/2.4.0+
Infernal/
Infernal/1.1.2
Mothur/
Mothur/1.33.3
Mothur/1.36.1
Mothur/1.38.1.1
""",
        )
        mapping_file = os.path.join(temp_directory, "mapping")
        with open(mapping_file, "w") as f:
            f.write("""
- from:
    name: blast+
    unversioned: true
  to:
    name: BlastPlus
    version: 2.4.0+
- from:
    name: Mothur
    version: 1.38
  to:
    version: 1.38.1.1
""")

        resolver = LmodDependencyResolver(
            _SimpleDependencyManager(),
            lmodexec=lmod_script,
            mapping_files=mapping_file,
            modulepath="/path/to/modulefiles",
        )

        lmod = resolver.resolve(ToolRequirement(name="BlastPlus", version="2.2.31+", type="package"))
        assert lmod.module_name == "BlastPlus"
        assert lmod.module_version == "2.2.31+", lmod.module_version

        lmod = resolver.resolve(ToolRequirement(name="blast+", type="package"))
        assert lmod.module_name == "BlastPlus"
        assert lmod.module_version == "2.4.0+", lmod.module_version

        lmod = resolver.resolve(ToolRequirement(name="blast+", version="2.23", type="package"))
        assert isinstance(lmod, NullDependency)

        lmod = resolver.resolve(ToolRequirement(name="Infernal", version="1.2", type="package"))
        assert isinstance(lmod, NullDependency)

        lmod = resolver.resolve(ToolRequirement(name="Mothur", version="1.38", type="package"))
        assert lmod.module_name == "Mothur"
        assert lmod.module_version == "1.38.1.1", lmod.module_version


def test_lmod_dependency_resolver_skip_availability_check():
    with __test_base_path() as temp_directory:
        lmod_script = _setup_lmod_command(
            temp_directory,
            """
/opt/apps/modulefiles:
BlastPlus/2.4.0+
Infernal/1.1.2
Mothur/1.36.1
""",
        )
        resolver = LmodDependencyResolver(
            _SimpleDependencyManager(),
            lmodexec=lmod_script,
            skip_availability_check="true",
            modulepath="/path/to/modulefiles",
        )

        lmod = resolver.resolve(ToolRequirement(name="Mothur", version="1.36.1", type="package"))
        assert lmod.module_name == "Mothur"
        assert lmod.module_version == "1.36.1"

        lmod = resolver.resolve(ToolRequirement(name="BlastPlus", version="2.3", type="package"))
        assert lmod.module_name == "BlastPlus"
        assert lmod.module_version == "2.3"

        lmod = resolver.resolve(ToolRequirement(name="Infernal", version=None, type="package"))
        assert lmod.module_name == "Infernal"
        assert lmod.module_version is None

        lmod = resolver.resolve(ToolRequirement(name="Foo", version="0.1", type="package"))
        assert lmod.module_name == "Foo"
        assert lmod.module_version == "0.1"

        lmod = resolver.resolve(ToolRequirement(name="Bar", version=None, type="package"))
        assert lmod.module_name == "Bar"
        assert lmod.module_version is None


def _setup_lmod_command(temp_directory, contents):
    lmod_script = os.path.join(temp_directory, "lmod")
    __write_script(
        lmod_script,
        f"""#!/bin/sh
cat {temp_directory}/lmod_example_output 1>&2;
""",
    )
    with open(os.path.join(temp_directory, "lmod_example_output"), "w") as f:
        # Subset of a "lmod -t avail" command of the LMOD environment module system.
        f.write(contents)
    return lmod_script


def test_lmod_dependency():
    with __test_base_path() as temp_directory:
        # Create mock lmod script that just exports a variable
        # the way "lmod load" would, but also validate correct
        # module name and version are coming through.
        mock_lmodexec = os.path.join(temp_directory, "pouet")
        __write_script(
            mock_lmodexec,
            """#!/bin/sh
if [ "$2" != "foomodule/1.0" ]; then
    exit 1
fi
echo 'FOO="bar"'
""",
        )
        resolver = Bunch(lmodexec=mock_lmodexec, settargexec=None, modulepath="/path/to/modulefiles")
        dependency = LmodDependency(resolver, "foomodule", "1.0")
        __assert_foo_exported(dependency.shell_commands())


def __write_script(path, contents):
    with open(path, "w") as f:
        f.write(contents)
    st = stat(path)
    chmod(path, st.st_mode | S_IXUSR)


def test_galaxy_dependency_object_script():
    with __test_base_path() as base_path:
        # Create env.sh file that just exports variable FOO and verify it
        # shell_commands export it correctly.
        env_path = __setup_galaxy_package_dep(base_path, TEST_REPO_NAME, TEST_VERSION, 'export FOO="bar"')
        dependency = GalaxyPackageDependency(
            env_path, os.path.dirname(env_path), TEST_REPO_NAME, "package", TEST_VERSION
        )
        __assert_foo_exported(dependency.shell_commands())


def test_shell_commands_built():
    # Test that dependency manager builds valid shell commands for a list of
    # requirements.
    with __test_base_path() as base_path:
        dm = __dependency_manager_for_base_path(default_base_path=base_path)
        __setup_galaxy_package_dep(base_path, TEST_REPO_NAME, TEST_VERSION, contents='export FOO="bar"')
        mock_requirements = ToolRequirements([{"type": "package", "version": TEST_VERSION, "name": TEST_REPO_NAME}])
        commands = dm.dependency_shell_commands(mock_requirements)
        __assert_foo_exported(commands)


def __assert_foo_exported(commands):
    command = ["bash", "-c", '{}; echo "$FOO"'.format("".join(commands))]
    process = Popen(command, stdout=PIPE)
    output = process.communicate()[0].strip()
    assert output == b"bar", f"Command {command} exports FOO as {unicodify(output)}, not bar"


def __setup_galaxy_package_dep(base_path, name, version, contents=""):
    dep_directory = os.path.join(base_path, name, version)
    env_path = os.path.join(dep_directory, "env.sh")
    __touch(env_path, contents)
    return env_path


def __touch(fname, data=None):
    dirname = os.path.dirname(fname)
    if not os.path.exists(dirname):
        makedirs(dirname)
    f = open(fname, "w")
    try:
        if data:
            f.write(data)
    finally:
        f.close()


def __build_test_repo(type, version=None):
    return Bunch(
        owner=TEST_REPO_USER,
        name=TEST_REPO_NAME,
        type=type,
        version=version,
        tool_shed_repository=Bunch(
            owner=TEST_REPO_USER, name=TEST_REPO_NAME, installed_changeset_revision=TEST_REPO_CHANGESET
        ),
    )


@contextmanager
def __test_base_path():
    base_path = tempfile.mkdtemp()
    try:
        yield base_path
    finally:
        rmtree(base_path)


def test_parse():
    with __parse_resolvers("""<dependency_resolvers>
  <tool_shed_packages />
  <galaxy_packages />
</dependency_resolvers>
""") as dependency_resolvers:
        assert "ToolShed" in dependency_resolvers[0].__class__.__name__
        assert "Galaxy" in dependency_resolvers[1].__class__.__name__

    with __parse_resolvers("""<dependency_resolvers>
  <galaxy_packages />
  <tool_shed_packages />
</dependency_resolvers>
""") as dependency_resolvers:
        assert "Galaxy" in dependency_resolvers[0].__class__.__name__
        assert "ToolShed" in dependency_resolvers[1].__class__.__name__

    with __parse_resolvers("""<dependency_resolvers>
  <galaxy_packages />
  <tool_shed_packages />
  <galaxy_packages versionless="true" />
</dependency_resolvers>
""") as dependency_resolvers:
        assert not dependency_resolvers[0].versionless
        assert dependency_resolvers[2].versionless

    with __parse_resolvers("""<dependency_resolvers>
  <galaxy_packages />
  <tool_shed_packages />
  <galaxy_packages base_path="/opt/galaxy/legacy/"/>
</dependency_resolvers>
""") as dependency_resolvers:
        # Unspecified base_paths are both default_base_paths
        assert dependency_resolvers[0].base_path == dependency_resolvers[1].base_path
        # Can specify custom base path...
        assert dependency_resolvers[2].base_path == "/opt/galaxy/legacy"
        # ... that is different from the default.
        assert dependency_resolvers[0].base_path != dependency_resolvers[2].base_path


def test_uses_tool_shed_dependencies():
    with __dependency_manager("""<dependency_resolvers>
  <galaxy_packages />
</dependency_resolvers>
""") as dm:
        assert not dm.uses_tool_shed_dependencies()

    with __dependency_manager("""<dependency_resolvers>
  <tool_shed_packages />
</dependency_resolvers>
""") as dm:
        assert dm.uses_tool_shed_dependencies()


def test_config_module_defaults():
    with __parse_resolvers("""<dependency_resolvers>
  <modules prefetch="false" />
</dependency_resolvers>
""") as dependency_resolvers:
        module_resolver = dependency_resolvers[0]
        assert module_resolver.module_checker.__class__.__name__ == "AvailModuleChecker"

    with __parse_resolvers(
        """
-  type: modules
   prefetch: false
""",
        extension=".yml",
    ) as dependency_resolvers:
        module_resolver = dependency_resolvers[0]
        assert module_resolver.module_checker.__class__.__name__ == "AvailModuleChecker"


def test_config_modulepath():
    # Test reads and splits MODULEPATH if modulepath is not specified.
    with __parse_resolvers("""<dependency_resolvers>
  <modules find_by="directory" modulepath="/opt/modules/modulefiles:/usr/local/modules/modulefiles" />
</dependency_resolvers>
""") as dependency_resolvers:
        assert dependency_resolvers[0].module_checker.directories == [
            "/opt/modules/modulefiles",
            "/usr/local/modules/modulefiles",
        ]


def test_config_MODULEPATH():
    # Test reads and splits MODULEPATH if modulepath is not specified.
    with modify_environ({"MODULEPATH": "/opt/modules/modulefiles:/usr/local/modules/modulefiles"}):
        with __parse_resolvers("""<dependency_resolvers>
  <modules find_by="directory" />
</dependency_resolvers>
""") as dependency_resolvers:
            assert dependency_resolvers[0].module_checker.directories == [
                "/opt/modules/modulefiles",
                "/usr/local/modules/modulefiles",
            ]


def test_config_MODULESHOME():
    # Test fallbacks to read MODULESHOME if modulepath is not specified and
    # neither is MODULEPATH.
    with modify_environ({"MODULESHOME": "/opt/modules"}, keys_to_remove=["MODULEPATH"]):
        with __parse_resolvers("""<dependency_resolvers>
  <modules find_by="directory" />
</dependency_resolvers>
""") as dependency_resolvers:
            assert dependency_resolvers[0].module_checker.directories == ["/opt/modules/modulefiles"]


def test_config_module_directory_searcher():
    with __parse_resolvers("""<dependency_resolvers>
  <modules find_by="directory" modulepath="/opt/Modules/modulefiles" />
</dependency_resolvers>
""") as dependency_resolvers:
        module_resolver = dependency_resolvers[0]
        assert module_resolver.module_checker.directories == ["/opt/Modules/modulefiles"]


def test_dependency_manager_config_options_global():
    app_config = {
        "tool_dependency_dir": "/tmp",
        "tool_dependency_cache_dir": "/tmp",
        "conda_auto_init": False,
    }
    dm = __dependency_manager_for_config(app_config.copy())
    assert not dm.to_dict()["cache"]

    app_config["use_cached_dependency_manager"] = True
    dm = __dependency_manager_for_config(app_config.copy())
    assert dm.to_dict()["cache"]
    assert dm.to_dict()["precache"]

    app_config["precache_dependencies"] = False
    dm = __dependency_manager_for_config(app_config.copy())
    assert not dm.to_dict()["precache"]

    conda_opts = _first_conda_resolver_options(dm)
    assert conda_opts["use_local"] is False

    app_config["conda_use_local"] = True
    dm = __dependency_manager_for_config(app_config.copy())
    conda_opts = _first_conda_resolver_options(dm)
    assert conda_opts["use_local"] is True


def test_dependency_manager_config_options_embedded_config():
    dependency_config: Dict[str, Any] = {
        "default_base_path": "/tmp",
        "cache_dir": "/tmp",
    }
    app_config = {
        "dependency_resolution": dependency_config,
        "conda_auto_init": False,
    }
    dm = __dependency_manager_for_config(app_config.copy())
    assert not dm.to_dict()["cache"]

    dependency_config["cache"] = True
    dm = __dependency_manager_for_config(app_config.copy())
    assert dm.to_dict()["cache"]
    assert dm.to_dict()["precache"]

    dependency_config["precache"] = False
    dm = __dependency_manager_for_config(app_config.copy())
    assert not dm.to_dict()["precache"]

    conda_opts = _first_conda_resolver_options(dm)
    assert conda_opts["use_local"] is False

    app_config["conda_use_local"] = True
    dm = __dependency_manager_for_config(app_config.copy())
    conda_opts = _first_conda_resolver_options(dm)
    assert conda_opts["use_local"] is True


def test_dependency_manager_config_options_resolution_config():
    app_config = {
        "conda_auto_init": False,
    }
    resolution_config: Dict[str, Any] = {
        "default_base_path": "/tmp",
        "cache_dir": "/tmp",
    }
    dm = __dependency_manager_for_config(app_config.copy(), resolution_config=resolution_config.copy())
    assert not dm.to_dict()["cache"]

    resolution_config["cache"] = True
    dm = __dependency_manager_for_config(app_config.copy(), resolution_config=resolution_config.copy())
    assert dm.to_dict()["cache"]
    assert dm.to_dict()["precache"]

    resolution_config["precache"] = False
    dm = __dependency_manager_for_config(app_config.copy(), resolution_config=resolution_config.copy())
    assert not dm.to_dict()["precache"]

    conda_opts = _first_conda_resolver_options(dm)
    assert conda_opts["use_local"] is False

    app_config["conda_use_local"] = True
    dm = __dependency_manager_for_config(app_config.copy(), resolution_config=resolution_config.copy())
    conda_opts = _first_conda_resolver_options(dm)
    assert conda_opts["use_local"] is True


def test_dependency_manager_none():
    # by default tool_dependency_dir will be use to create some default resolvers...
    app_config = {
        "conda_auto_init": False,
        "tool_dependency_dir": "some_not_none_value",
    }
    dm = __dependency_manager_for_config(app_config.copy())
    assert dm.to_dict()["use"]

    # but setting it none disables dependency resolution unless explicit resolvers
    # are configured.
    app_config = {
        "conda_auto_init": False,
        "tool_dependency_dir": "none",
    }
    dm = __dependency_manager_for_config(app_config.copy())
    assert not dm.to_dict()["use"]


def _first_conda_resolver_options(dm):
    return [r for r in dm.to_dict()["resolvers"] if r["resolver_type"] == "conda"][0]


@contextmanager
def __parse_resolvers(file_content, extension=".xml"):
    with __dependency_manager(file_content, extension=extension) as dm:
        yield dm.dependency_resolvers


@contextmanager
def __dependency_manager(file_content, extension=".xml"):
    with __test_base_path() as base_path:
        with tempfile.NamedTemporaryFile("w+", suffix=extension) as tmp:
            tmp.write(file_content)
            tmp.flush()
            dm = __dependency_manager_for_base_path(default_base_path=base_path, conf_file=tmp.name)
            yield dm


def __dependency_manager_for_base_path(default_base_path: str, conf_file: Optional[str] = None) -> DependencyManager:
    dm = DependencyManager(
        default_base_path=default_base_path, conf_file=conf_file, app_config={"conda_auto_init": False}
    )
    if ROUND_TRIP_TEST_DEPENDENCY_MANAGER_SERIALIZATION:
        as_dict = dm.to_dict()
        dm = build_dependency_manager(resolution_config_dict=as_dict)
    return dm


def __dependency_manager_for_config(app_config, resolution_config=None):
    dm = build_dependency_manager(app_config_dict=app_config, resolution_config_dict=resolution_config)
    if ROUND_TRIP_TEST_DEPENDENCY_MANAGER_SERIALIZATION:
        as_dict = dm.to_dict()
        dm = build_dependency_manager(resolution_config_dict=as_dict)
    return dm


class _SimpleDependencyManager:
    default_base_path = None

    def get_resolver_option(self, resolver, key, explicit_resolver_options=None):
        if explicit_resolver_options is None:
            explicit_resolver_options = {}
        return explicit_resolver_options.get(key)



from galaxy.tool_util.parser.factory import get_tool_source
from galaxy.tool_util.parser.output_objects import from_tool_source
from galaxy.tool_util.unittest_utils import functional_test_tool_path


def test_from_tool_data_table():
    tool_source = get_tool_source(functional_test_tool_path("dbkey_output_action.xml"))
    # prevent regression of https://github.com/galaxyproject/galaxy/issues/18554. Tool fails without fix
    from_tool_source(tool_source)



from galaxy.tool_util.parameters.convert import cwl_runtime_model
from galaxy.tool_util.unittest_utils.parameters import parameter_bundle_for_file


def test_cwl_runtime_model_conditional():
    model = parameter_bundle_for_file("gx_conditional_select")
    rval = cwl_runtime_model(model)
    assert rval


def test_cwl_runtime_model_data():
    model = parameter_bundle_for_file("gx_data")
    rval = cwl_runtime_model(model)
    assert rval


def test_cwl_runtime_model_data_collection_list():
    model = parameter_bundle_for_file("gx_data_collection_list")
    rval = cwl_runtime_model(model)
    assert rval


def test_cwl_runtime_model_data_collection_unknown_type():
    model = parameter_bundle_for_file("gx_data_collection")
    rval = cwl_runtime_model(model)
    assert rval



LOC_ALPHA_CONTENTS_V2 = """
data1	data1name	${__HERE__}/data1/entry.txt
data2	data2name	${__HERE__}/data2/entry.txt
data3	data3name	${__HERE__}/data3/entry.txt
"""


def test_data_tables_as_dictionary(tdt_manager):
    assert "testalpha" in tdt_manager.data_tables
    assert "testdelta" not in tdt_manager.data_tables


def test_to_dict(tdt_manager):
    as_dict = tdt_manager.to_dict()
    assert "testalpha" in as_dict
    assert "testdelta" not in as_dict
    testalpha_as_dict = as_dict["testalpha"]
    assert "columns" in testalpha_as_dict


def test_index(tdt_manager):
    index = tdt_manager.index()
    assert len(index.root) >= 1
    entry = index.find_entry("testalpha")
    assert entry
    entry = index.find_entry("testomega")
    assert not entry


def test_reload(tdt_manager, tmp_path):
    assert len(tdt_manager["testalpha"].data) == 2
    loc1 = tmp_path / "testalpha.loc"
    loc1.write_text(LOC_ALPHA_CONTENTS_V2)
    tdt_manager.reload_tables()
    assert len(tdt_manager["testalpha"].data) == 3


def test_reload_by_path(tdt_manager, tmp_path):
    assert len(tdt_manager["testalpha"].data) == 2
    loc1 = tmp_path / "testalpha.loc"
    loc1.write_text(LOC_ALPHA_CONTENTS_V2)
    tdt_manager.reload_tables(path=str(loc1))
    assert len(tdt_manager["testalpha"].data) == 3


def test_reload_by_name(tdt_manager, tmp_path):
    assert len(tdt_manager["testalpha"].data) == 2
    loc1 = tmp_path / "testalpha.loc"
    loc1.write_text(LOC_ALPHA_CONTENTS_V2)
    tdt_manager.reload_tables("testalpha")
    assert len(tdt_manager["testalpha"].data) == 3


def test_merging_tables(merged_tdt_manager):
    assert len(merged_tdt_manager["testbeta"].data) == 2


def test_to_json(merged_tdt_manager, tmp_path):
    json_path = tmp_path / "as_json.json"
    assert not json_path.exists()
    merged_tdt_manager.to_json(json_path)
    assert json_path.exists()



import os
import re
from typing import (
    Any,
    List,
    Optional,
    Tuple,
)

from galaxy.tool_util.model_factory import parse_tool
from galaxy.tool_util.parameters import (
    DataCollectionRequest,
    DataRequestHda,
    encode_test,
    input_models_for_tool_source,
)
from galaxy.tool_util.parameters.case import (
    test_case_state as case_state,
    TestCaseStateAndWarnings,
    TestCaseStateValidationResult,
    validate_test_cases_for_tool_source,
)
from galaxy.tool_util.parser.factory import get_tool_source
from galaxy.tool_util.parser.interface import (
    ToolSource,
    ToolSourceTest,
)
from galaxy.tool_util.unittest_utils import functional_test_tool_directory
from galaxy.tool_util.verify.parse import parse_tool_test_descriptions
from galaxy.tool_util_models.tool_source import (
    JsonTestCollectionDefDict,
    JsonTestDatasetDefDict,
)
from .util import dict_verify_each

# legacy tools allows specifying parameter and repeat parameters without
# qualification. This was problematic and could result in ambigious specifications.
TOOLS_THAT_USE_UNQUALIFIED_PARAMETER_ACCESS = [
    "boolean_conditional.xml",
    "simple_constructs.xml",
    "section.xml",
    "collection_paired_conditional_structured_like.xml",
    "output_action_change_format.xml",
    "top_level_data.xml",
    "disambiguate_cond.xml",
    "multi_repeats.xml",
    "implicit_default_conds.xml",
    "min_repeat.xml",
]

# tools that use truevalue/falsevalue in parameter setting, I think we're going to
# forbid this for a future tool profile version. Potential ambigouity could result.
TOOLS_THAT_USE_TRUE_FALSE_VALUE_BOOLEAN_SPECIFICATION = [
    "inputs_as_json_profile.xml",
    "inputs_as_json_with_paths.xml",
    "inputs_as_json.xml",
]

TOOLS_THAT_USE_SELECT_BY_VALUE = [
    "multi_select.xml",
]

# Figure out the problem and resolve.
TOOLS_THAT_ARE_OUTSTANDING_ISSUES = [
    "gx_conditional_boolean_optional.xml",
    "gx_conditional_boolean_discriminate_on_string_value.xml",
]

TEST_TOOL_THAT_DO_NOT_VALIDATE = (
    TOOLS_THAT_USE_UNQUALIFIED_PARAMETER_ACCESS
    + TOOLS_THAT_USE_TRUE_FALSE_VALUE_BOOLEAN_SPECIFICATION
    + TOOLS_THAT_USE_SELECT_BY_VALUE
    + TOOLS_THAT_ARE_OUTSTANDING_ISSUES
    + [
        # will never handle upload_dataset
        "upload.xml",
    ]
)

MOCK_ID = "thisisafakeid"


def test_parameter_test_cases_validate():
    validation_result = validate_test_cases_for("column_param")
    assert len(validation_result[0].warnings) == 0
    assert len(validation_result[1].warnings) == 0
    assert len(validation_result[2].warnings) == 1

    validation_result = validate_test_cases_for("column_param", use_latest_profile=True)
    assert validation_result[2].validation_error


def test_legacy_features_fail_validation_with_24_2(tmp_path):
    for filename in TOOLS_THAT_USE_UNQUALIFIED_PARAMETER_ACCESS + TOOLS_THAT_USE_TRUE_FALSE_VALUE_BOOLEAN_SPECIFICATION:
        _assert_tool_test_parsing_only_fails_with_newer_profile(tmp_path, filename, index=None)

    # column parameters need to be indexes
    _assert_tool_test_parsing_only_fails_with_newer_profile(tmp_path, "column_param.xml", index=2)

    # selection by value only
    _assert_tool_test_parsing_only_fails_with_newer_profile(tmp_path, "multi_select.xml", index=1)


def _assert_tool_test_parsing_only_fails_with_newer_profile(tmp_path, filename: str, index: Optional[int] = 0):
    test_tool_directory = functional_test_tool_directory()
    original_path = os.path.join(test_tool_directory, filename)
    new_path = tmp_path / filename
    with open(original_path) as rf:
        tool_contents = rf.read()
        tool_contents = re.sub(r'profile="[\d\.]*"', r"", tool_contents)
        new_profile_contents = tool_contents.replace("<tool ", '<tool profile="24.2" ', 1)
    with open(new_path, "w") as wf:
        wf.write(new_profile_contents)
    test_cases = list(parse_tool_test_descriptions(get_tool_source(original_path)))
    if index is not None:
        assert test_cases[index].to_dict()["error"] is False
    else:
        # just make sure there is at least one failure...
        assert not any(c.to_dict()["error"] is True for c in test_cases)

    test_cases = list(parse_tool_test_descriptions(get_tool_source(new_path)))
    if index is not None:
        assert (
            test_cases[index].to_dict()["error"] is True
        ), f"expected {filename} to have validation failure preventing loading of tools"
    else:
        assert any(c.to_dict()["error"] is True for c in test_cases)


def test_validate_framework_test_tools():
    test_tool_directory = functional_test_tool_directory()
    parameter_tool_directory = os.path.join(test_tool_directory, "parameters")
    for test_directory in [test_tool_directory, parameter_tool_directory]:
        for tool_name in os.listdir(test_directory):
            if tool_name in TEST_TOOL_THAT_DO_NOT_VALIDATE:
                continue
            if tool_name.endswith("_conf.xml") or tool_name == "macros.xml":
                # tool conf (toolbox) files or sample datatypes
                continue
            tool_path = os.path.join(test_directory, tool_name)
            if not tool_path.endswith(".xml") or os.path.isdir(tool_path):
                continue

            try:
                _validate_path(tool_path)
            except Exception as e:
                raise Exception(f"Failed to validate {tool_path}: {str(e)}")


def test_test_case_state_conversion():
    tool_source = tool_source_for("collection_nested_test")
    test_cases: List[ToolSourceTest] = tool_source.parse_tests_to_dict()["tests"]
    state = case_state_for(tool_source, test_cases[0])
    expectations: List[Tuple[List[Any], Optional[Any]]]
    expectations = [
        (["f1", "collection_type"], "list:paired"),
        (["f1", "class"], "Collection"),
        (["f1", "elements", 0, "class"], "Collection"),
        (["f1", "elements", 0, "collection_type"], "paired"),
        (["f1", "elements", 0, "elements", 0, "class"], "File"),
        (["f1", "elements", 0, "elements", 0, "path"], "simple_line.txt"),
        (["f1", "elements", 0, "elements", 0, "identifier"], "forward"),
    ]
    dict_verify_each(state.tool_state.input_state, expectations)

    tool_source = tool_source_for("dbkey_filter_input")
    test_cases = tool_source.parse_tests_to_dict()["tests"]
    state = case_state_for(tool_source, test_cases[0])
    expectations = [
        (["inputs", "class"], "File"),
        (["inputs", "dbkey"], "hg19"),
    ]
    dict_verify_each(state.tool_state.input_state, expectations)

    tool_source = tool_source_for("discover_metadata_files")
    test_cases = tool_source.parse_tests_to_dict()["tests"]
    state = case_state_for(tool_source, test_cases[0])
    expectations = [
        (["input_bam", "class"], "File"),
        (["input_bam", "filetype"], "bam"),
    ]
    dict_verify_each(state.tool_state.input_state, expectations)

    tool_source = tool_source_for("remote_test_data_location")
    test_cases = tool_source.parse_tests_to_dict()["tests"]
    state = case_state_for(tool_source, test_cases[0])
    expectations = [
        (["input", "class"], "File"),
        (
            ["input", "location"],
            "https://raw.githubusercontent.com/galaxyproject/planemo/7be1bf5b3971a43eaa73f483125bfb8cabf1c440/tests/data/hello.txt",
        ),
    ]
    dict_verify_each(state.tool_state.input_state, expectations)

    tool_source = tool_source_for("composite")
    test_cases = tool_source.parse_tests_to_dict()["tests"]
    state = case_state_for(tool_source, test_cases[0])
    expectations = [
        (["input", "class"], "File"),
        (["input", "filetype"], "velvet"),
        (["input", "composite_data", 0], "velveth_test1/Sequences"),
    ]
    dict_verify_each(state.tool_state.input_state, expectations)

    tool_source = tool_source_for("parameters/gx_group_tag")
    test_cases = tool_source.parse_tests_to_dict()["tests"]
    state = case_state_for(tool_source, test_cases[0])
    expectations = [
        (["ref_parameter", "class"], "Collection"),
        (["ref_parameter", "collection_type"], "paired"),
        (["ref_parameter", "elements", 0, "identifier"], "forward"),
        (["ref_parameter", "elements", 0, "tags", 0], "group:type:single"),
    ]
    dict_verify_each(state.tool_state.input_state, expectations)

    index = 2
    tool_source = tool_source_for("filter_param_value_ref_attribute")
    test_cases = tool_source.parse_tests_to_dict()["tests"]
    state = case_state_for(tool_source, test_cases[index])
    expectations = [
        (["data_mult", 0, "path"], "1.bed"),
        (["data_mult", 0, "dbkey"], "hg19"),
        (["data_mult", 1, "path"], "2.bed"),
        (["data_mult", 0, "dbkey"], "hg19"),
    ]
    dict_verify_each(state.tool_state.input_state, expectations)

    index = 1
    tool_source = tool_source_for("expression_pick_larger_file")
    test_cases = tool_source.parse_tests_to_dict()["tests"]
    state = case_state_for(tool_source, test_cases[index])
    expectations = [
        (["input1", "path"], "simple_line_alternative.txt"),
        (["input2"], None),
    ]
    dict_verify_each(state.tool_state.input_state, expectations)

    index = 2
    state = case_state_for(tool_source, test_cases[index])
    expectations = [
        (["input1"], None),
        (["input2", "path"], "simple_line.txt"),
    ]
    dict_verify_each(state.tool_state.input_state, expectations)

    index = 0
    tool_source = tool_source_for("composite_shapefile")
    test_cases = tool_source.parse_tests_to_dict()["tests"]
    state = case_state_for(tool_source, test_cases[index])
    expectations = [
        (["input", "filetype"], "shp"),
        (["input", "composite_data", 0], "shapefile/shapefile.shp"),
    ]
    dict_verify_each(state.tool_state.input_state, expectations)


def test_convert_to_requests():
    tools = [
        "parameters/gx_drill_down_recurse_multiple",
        "parameters/gx_conditional_select",
        "expression_pick_larger_file",
        "identifier_in_conditional",
        "column_param_list",
        "composite_shapefile",
    ]
    for tool_path in tools:
        tool_source = tool_source_for(tool_path)
        parameters = input_models_for_tool_source(tool_source)
        parsed_tool = parse_tool(tool_source)
        profile = tool_source.parse_profile()
        test_cases: List[ToolSourceTest] = tool_source.parse_tests_to_dict()["tests"]

        def mock_adapt_datasets(input: JsonTestDatasetDefDict) -> DataRequestHda:
            return DataRequestHda(src="hda", id=MOCK_ID)

        def mock_adapt_collections(input: JsonTestCollectionDefDict) -> DataCollectionRequest:
            return DataCollectionRequest(src="hdca", id=MOCK_ID)

        for test_case in test_cases:
            if test_case.get("expect_failure"):
                continue
            test_case_state_and_warnings = case_state(test_case, parsed_tool.inputs, profile)
            test_case_state = test_case_state_and_warnings.tool_state

            encode_test(test_case_state, parameters, mock_adapt_datasets, mock_adapt_collections)


def _validate_path(tool_path: str):
    tool_source = get_tool_source(tool_path)
    tool_id = tool_source.parse_id()
    model_name = f"{tool_id} (test case model)"
    parsed_tool = parse_tool(tool_source)
    profile = tool_source.parse_profile()
    test_cases: List[ToolSourceTest] = tool_source.parse_tests_to_dict()["tests"]
    for test_case in test_cases:
        if test_case.get("expect_failure"):
            continue
        test_case_state_and_warnings = case_state(test_case, parsed_tool.inputs, profile, name=model_name)
        tool_state = test_case_state_and_warnings.tool_state
        assert tool_state.state_representation == "test_case_xml"


def validate_test_cases_for(tool_name: str, **kwd) -> List[TestCaseStateValidationResult]:
    return validate_test_cases_for_tool_source(tool_source_for(tool_name), **kwd)


def case_state_for(tool_source: ToolSource, test_case: ToolSourceTest) -> TestCaseStateAndWarnings:
    parsed_tool = parse_tool(tool_source)
    profile = tool_source.parse_profile()
    return case_state(test_case, parsed_tool.inputs, profile)


def tool_source_for(tool_name: str) -> ToolSource:
    test_tool_directory = functional_test_tool_directory()
    tool_path = os.path.join(test_tool_directory, f"{tool_name}.xml")
    tool_source = get_tool_source(tool_path)
    return tool_source



from galaxy.tool_util.biotools.source import GitContentBiotoolsMetadataSource
from ._util import content_dir


def test_edam_parse_bwa():
    metadata_source = GitContentBiotoolsMetadataSource(content_dir)
    biotools_entry = metadata_source.get_biotools_metadata("bwa")
    assert biotools_entry
    parsed_edam_info = biotools_entry.edam_info
    assert len(parsed_edam_info.edam_operations) == 4
    assert "operation_3198" in parsed_edam_info.edam_operations
    assert len(parsed_edam_info.edam_topics) == 1
    assert "topic_0102" in parsed_edam_info.edam_topics



import pytest

from galaxy.tool_util.data import ToolDataTableManager

LOC_ALPHA_CONTENTS = """
data1	data1name	${__HERE__}/data1/entry.txt
data2	data2name	${__HERE__}/data2/entry.txt
"""


LOC_BETA_CONTENTS_1 = """
beta1	beta1name	${__HERE__}/beta1/entry.txt
"""


LOC_BETA_CONTENTS_2 = """
beta2	beta2name	${__HERE__}/beta2/entry.txt
"""


TOOL_DATA_TABLE_CONF_XML = """<tables>
  <table name="testalpha" comment_char="#">
    <columns>value, name, path</columns>
    <file path="${__HERE__}/testalpha.loc" />
  </table>
</tables>
"""


MERGED_TOOL_DATA_TABLE_CONF_XML_1 = """<tables>
  <table name="testbeta" comment_char="#">
    <columns>value, name, path</columns>
    <file path="${__HERE__}/testbeta1.loc" />
  </table>
</tables>
"""


MERGED_TOOL_DATA_TABLE_CONF_XML_2 = """<tables>
  <table name="testbeta" comment_char="#">
    <columns>value, name, path</columns>
    <file path="${__HERE__}/testbeta2.loc" />
  </table>
</tables>
"""


@pytest.fixture
def tdt_manager(tmp_path) -> ToolDataTableManager:
    _write_loc_files(tmp_path)
    conf = tmp_path / "tool_data_table_conf.xml"
    conf.write_text(TOOL_DATA_TABLE_CONF_XML)
    return ToolDataTableManager(tmp_path, conf)


@pytest.fixture
def merged_tdt_manager(tmp_path) -> ToolDataTableManager:
    _write_loc_files(tmp_path)
    conf1 = tmp_path / "tool_data_table_conf_1.xml"
    conf1.write_text(MERGED_TOOL_DATA_TABLE_CONF_XML_1)
    conf2 = tmp_path / "tool_data_table_conf_2.xml"
    conf2.write_text(MERGED_TOOL_DATA_TABLE_CONF_XML_2)
    return ToolDataTableManager(tmp_path, f"{conf1},{conf2}")


def _write_loc_files(tmp_path):
    loc1 = tmp_path / "testalpha.loc"
    loc1.write_text(LOC_ALPHA_CONTENTS)

    loc2 = tmp_path / "testbeta1.loc"
    loc2.write_text(LOC_BETA_CONTENTS_1)

    loc3 = tmp_path / "testbeta2.loc"
    loc3.write_text(LOC_BETA_CONTENTS_2)

    data1 = tmp_path / "data1"
    data1.mkdir()
    data1_entry = data1 / "entry.txt"
    data1_entry.write_text("This is data 1.")

    data2 = tmp_path / "data2"
    data2.mkdir()
    data2_entry = data2 / "entry.txt"
    data2_entry.write_text("This is data 2.")

    data3 = tmp_path / "data3"
    data3.mkdir()
    data3_entry = data3 / "entry.txt"
    data3_entry.write_text("This is data 3.")

    beta1 = tmp_path / "beta1"
    beta1.mkdir()
    beta1_entry = beta1 / "entry.txt"
    beta1_entry.write_text("This is beta 1.")

    beta2 = tmp_path / "beta2"
    beta2.mkdir()
    beta2_entry = beta2 / "entry.txt"
    beta2_entry.write_text("This is beta 2.")



from string import Template

import lxml.etree as ET
import pytest
from pydantic import ValidationError

from galaxy.tool_util.verify.codegen import galaxy_xsd_path
from galaxy.tool_util.verify.parse import assertion_xml_els_to_models
from galaxy.tool_util_models.assertions import assertion_list
from galaxy.util.commands import shell
from galaxy.util.unittest_utils import skip_unless_executable

valid_assertions = [
    {"that": "has_size", "size": "5G"},
    {"that": "has_size", "size": 1},
    {"that": "has_size", "size": "5Mi"},
    {"that": "has_text", "text": "JBrowseDefaultMainPage"},
    {"that": "has_line", "line": "'>Wildtype Staphylococcus aureus strain WT.'"},
    {"that": "has_n_columns", "n": 2},
    {"that": "is_valid_xml"},
    {"that": "has_element_with_path", "path": "//el"},
    {"that": "has_n_elements_with_path", "path": "//el", "n": 4},
    {"that": "element_text_matches", "expression": "foob[a]r", "path": "//el"},
    {"that": "element_text_is", "text": "foobar", "path": "//el"},
    {
        "that": "xml_element",
        "path": "./elem/more[2]",
        "children": [{"that": "has_text_matching", "expression": "foo$"}],
    },
    {"that": "xml_element", "path": "./elem/more[2]"},
    {"that": "element_text", "path": "./elem/more[2]", "children": [{"that": "has_text", "text": "foo"}]},
    {"that": "has_json_property_with_value", "property": "foobar", "value": "'6'"},
    {"that": "has_json_property_with_text", "property": "foobar", "text": "cowdog"},
    {"that": "has_archive_member", "path": ".*/my-file.txt"},
    {"that": "has_archive_member", "path": ".*/my-file.txt", "children": [{"that": "has_text", "text": "1235abc"}]},
    {"that": "has_image_width", "width": 560},
    {"that": "has_image_width", "width": 560, "delta": 490},
    {"that": "has_image_height", "height": 560, "delta": 490},
    {"that": "has_image_height", "min": 45, "max": 90},
    {"that": "has_image_channels", "channels": 3},
    {"that": "has_image_channels", "channels": 3, "delta": 1},
    {"that": "has_image_channels", "min": 1, "max": 4},
    {"that": "has_image_channels", "min": 1, "max": 4, "negate": True},
    {"that": "has_image_mean_intensity", "mean_intensity": 3.4},
    {"that": "has_image_mean_intensity", "mean_intensity": 3.4, "eps": 0.2},
    {"that": "has_image_mean_intensity", "mean_intensity": 3.4, "eps": 0.2, "channel": 1},
    {"that": "has_image_mean_intensity", "min": 0.4, "max": 0.6, "channel": 1},
    {"that": "has_image_center_of_mass", "center_of_mass": "511.07, 223.34"},
    {"that": "has_image_center_of_mass", "center_of_mass": "511.07, 223.34", "channel": 1},
    {"that": "has_image_center_of_mass", "center_of_mass": "511.07, 223.34", "channel": 1, "eps": 0.2},
    {"that": "has_image_n_labels", "n": 85},
    {"that": "has_image_n_labels", "labels": [1, 3, 4]},
    {"that": "has_image_n_labels", "n": 9, "exclude_labels": [1, 3, 4]},
    {"that": "has_image_n_labels", "n": 9, "exclude_labels": [1, 3, 4], "negate": True},
    {"that": "has_image_mean_object_size", "mean_object_size": 9, "exclude_labels": [1, 3, 4]},
    {"that": "has_image_mean_object_size", "mean_object_size": 9, "labels": [1, 3, 4]},
    {"that": "has_image_mean_object_size", "mean_object_size": 9, "channel": 1, "eps": 0.2},
    {"that": "has_json_property_with_value", "property": "skipped_columns", "value": "[1, 3, 5]"},
    {"that": "has_json_property_with_text", "property": "color", "text": "red"},
    {"that": "has_n_columns", "n": 30},
    {"that": "has_n_columns", "n": 30, "delta": 4},
    {"that": "has_n_columns", "n": 30, "delta": 4, "sep": " ", "comment": "####"},
]

valid_xml_assertions = [
    """<has_size value="5G" />""",
    """<has_size size="5G" />""",
    """<has_size size="5G" />""",
    """<has_text text="JBrowseDefaultMainPage" />""",
    """<has_line line="'>Wildtype Staphylococcus aureus strain WT.'" />""",
    """<xml_element path="./elem/more[2]" attribute="name"><has_text_matching expression="foo$"/></xml_element>""",
    """<xml_element path="./elem/more[2]" attribute="name"></xml_element>""",
    """<element_text path="./elem/more[2]"><has_text text="foo" /></element_text>""",
    """<has_element_with_path path="BlastOutput_param/Parameters/Parameters_matrix" />""",
    r"""<element_text_matches path="BlastOutput_version" expression="BLASTP\s+2\.2.*"/>""",
    """<element_text_is path="BlastOutput_program" text="blastp" />""",
    r"""<attribute_matches path="outerElement/innerElement2" attribute="foo2" expression="bar\d+" />""",
    """<element_text path="BlastOutput_iterations/Iteration/Iteration_hits/Hit/Hit_def"><not_has_text text="EDK72998.1" /></element_text>""",
    """<has_archive_member path=".*/my-file.txt"><not_has_text text="EDK72998.1"/></has_archive_member>""",
    """<has_archive_member path=".*/my-file.txt"></has_archive_member>""",
    """<has_image_n_labels n="85" />""",
    """<has_image_center_of_mass center_of_mass="511.07, 223.34" />""",
    """<has_image_center_of_mass center_of_mass="511.07, 223.34" eps=".2" />""",
    """<has_image_mean_intensity mean_intensity="0.83" />""",
    """<has_image_height height="512" delta="2" />""",
    """<has_image_width width="512" delta="2" />""",
    """<has_image_channels channels="3" />""",
    """<has_image_channels channels="3" delta="1" />""",
    """<has_image_channels min="3" max="5" />""",
    """<has_image_channels min="3" max="5" negate="true" />""",
    """<has_json_property_with_value property="skipped_columns" value="[1, 3, 5]" />""",
    """<has_json_property_with_text property="color" text="red" />""",
    """<has_n_columns n="30" />""",
    """<has_n_columns n="30" delta="4" />""",
    """<has_n_columns n="30" delta="4" sep=" " comment="###" />""",
    """<has_image_width min="500" />""",
    """<has_image_height min="500" />""",
]

invalid_assertions = [
    {"that": "has_size", "size": "5Gigabytes"},
    # negative sizes not allowed
    {"that": "has_size", "size": -1},
    {"that": "has_n_columns", "n": [2]},
    {"that": "has_n_columns", "n": -2},
    {"that": "is_valid_xml_foo"},
    {"that": "has_element_with_path", "path": 45},
    {"that": "has_n_elements_with_path", "n": 4},
    {"that": "has_n_elements_with_path", "n": -4},
    {"that": "element_text_matches", "expression": 12, "path": "//el"},
    # unclosed regex group
    {"that": "element_text_matches", "expression": "[12", "path": "//el"},
    {"that": "xml_element", "path": "./elem/more[2]", "children": [{"that": "foobar"}]},
    {
        "that": "xml_element",
        "path": "./elem/more[2]",
        "children": [{"that": "has_text_matching", "line": "invalidprop"}],
    },
    # must specify children for element_text
    {"that": "element_text", "path": "./elem/more[2]"},
    {"that": "has_json_property_with_value", "property": 42, "value": "cowdog"},
    {"that": "has_json_property_with_text", "property": "foobar", "text": 6},
    {"that": "has_archive_member", "path": ".*/my-file.txt", "extra": "param"},
    {"that": "has_archive_member", "path": ".*/my-file.txt", "children": [{"that": "invalid"}]},
    {"that": "has_image_width", "width": "560"},
    {"that": "has_image_width", "width": -560},
    {"that": "has_image_width", "width": 560, "delta": "wrong"},
    {"that": "has_image_height", "height": -560},
    {"that": "has_image_center_of_mass", "center_of_mass": "511.07, 223.34, foobar"},
    {"that": "has_image_center_of_mass", "center_of_mass": "511.07"},
    {"that": "has_image_center_of_mass", "center_of_mass": "511.07, cow"},
    # negative mean object sizes are not allowed
    {"that": "has_image_n_labels", "n": -85},
    {"that": "has_image_n_labels", "n": 85, "delta": -3},
    # negative mean object sizes are not allowed
    {"that": "has_image_mean_object_size", "mean_object_size": -9, "exclude_labels": [1, 3, 4]},
    {"that": "has_image_mean_object_size", "mean_object_size": -9.0, "exclude_labels": [1, 3, 4]},
    {"that": "has_image_mean_object_size", "mean_object_size": 9, "exclude_labels": [1, 3, 4], "eps": -0.2},
    # looks a little odd in JSON but value is JSON loaded so must be a string
    {"that": "has_json_property_with_value", "property": "skipped_columns", "value": [1, 3, 5]},
    # missing property
    {"that": "has_json_property_with_value", "property": "skipped_columns"},
    {"that": "has_json_property_with_text", "property": "color"},
    {"that": "has_n_columns", "n": 30, "delta": "wrongtype"},
    {"that": "has_n_columns", "n": 30, "delta": -2},
    {"that": "has_n_columns", "n": 30, "delta": 4, "sep": " ", "comment": "####", "extra": "param"},
]

invalid_xml_assertions = [
    """<has_n_columns n="cow" />""",
    """<has_text line="JBrowseDefaultMainPage" />""",
    """<has_line text="JBrowseDefaultMainPage" />""",
    # at least one child assertion is required here...
    """<element_text path="./elem/more[2]"></element_text>""",
    """<has_archive_member path=".*/my-file.txt"><not_has_text line="EDK72998.1"/></has_archive_member>""",
    """<has_archive_member path=".*/my-file.txt"><foobar /></has_archive_member>""",
    """<has_image_center_of_mass center_of_mass="511.07, 223.34" extra="param" />""",
    """<has_image_center_of_mass center_of_mass="511.07, 223.34" eps="foobar" />""",
    # negative numbers not allowed
    """<has_image_n_labels n="-85" />""",
    """<has_image_n_labels n="85" delta="-3" />""",
    """<has_image_mean_intensity mean_intensity="not_a_number" />""",
    """<has_image_mean_intensity mean_intensity=".83" extra="param" />""",
    """<has_image_height height="512" delta="notanumber" />""",
    """<has_image_width width="512" delta="2" extra="param" />""",
    """<has_image_height height="512" delta="-2" />""",
    """<has_image_width width="512" delta="-2" />""",
    """<has_image_height height="-512" />""",
    """<has_image_width width="-512" />""",
    """<has_image_channels channels="3" exclude="wrong" />""",
    """<has_image_channels channels="3" delta="not_a_number"  />""",
    """<has_image_channels min="3" max="5" foo="bar" />""",
    """<has_image_channels min="3" max="5" negate="NOTABOOLEAN" />""",
    # missing required arguments...
    """<has_json_property_with_value property="skipped_columns" />""",
    """<has_json_property_with_text property="color" />""",
    """<has_n_columns n="30" delta="wrong_type" />""",
    """<has_n_columns n="30" delta="4" sep=" " comment="###" extra="param"/>""",
]


TOOL_TEMPLATE = Template("""
<tool id="gx_test" name="gx_test" version="1.0.0">
    <command><![CDATA[
echo '$parameter' >> '$output'
    ]]></command>
    <inputs>
        <param name="parameter" value="1" type="integer" />
    </inputs>
    <outputs>
        <data name="output" format="txt" />
    </outputs>
    <tests>
        <test>
            <param name="parameter" value_json="12456" />
            <output name="output">
                <assert_contents>
                    $assertion_xml
                </assert_contents>
            </output>
        </test>
    </tests>
</tool>
""")


def test_valid_json_models_validate():
    assertion_list.model_validate(valid_assertions)


def test_invalid_json_models_do_not_validate():
    for invalid_assertion in invalid_assertions:
        with pytest.raises(ValidationError):
            assertion_list.model_validate([invalid_assertion])


@skip_unless_executable("xmllint")
def test_valid_xsd(tmp_path):
    for assertion_xml in valid_xml_assertions:
        tool_xml = TOOL_TEMPLATE.safe_substitute(assertion_xml=assertion_xml)
        tool_path = tmp_path / "tool.xml"
        tool_path.write_text(tool_xml)
        ret = shell(["xmllint", "--nowarning", "--noout", "--schema", galaxy_xsd_path, str(tool_path)])
        assert ret == 0, f"{assertion_xml} failed to validate"


@skip_unless_executable("xmllint")
def test_invalid_xsd(tmp_path):
    for assertion_xml in invalid_xml_assertions:
        tool_xml = TOOL_TEMPLATE.safe_substitute(assertion_xml=assertion_xml)
        tool_path = tmp_path / "tool.xml"
        tool_path.write_text(tool_xml)
        ret = shell(["xmllint", "--nowarning", "--noout", "--schema", galaxy_xsd_path, str(tool_path)])
        assert ret != 0, f"{assertion_xml} validated when error expected"


def test_valid_xml_models_validate_after_json_transform():
    for assertion_xml in valid_xml_assertions:
        assertion_xml_els_to_models([ET.fromstring(assertion_xml)])



from typing import (
    Callable,
    Dict,
    Optional,
)

from galaxy.tool_util.unittest_utils import t_data_downloader_for
from galaxy.tool_util.verify.interactor import verify_hid


def dataset_fetcher_for(
    expected_hda_id: str, content: Dict[Optional[str], bytes]
) -> Callable[[str, Optional[str]], bytes]:
    def get_content(hda_id, filename: Optional[str] = None) -> bytes:
        assert expected_hda_id == hda_id
        return content[filename]

    return get_content


TEST_ID = "mydatasetid"


def test_verify_hid_simple_pass():
    test_data_downloader = t_data_downloader_for(b"expected")
    dataset_fetcher = dataset_fetcher_for(TEST_ID, {None: b"expected"})
    verify_hid(
        "",
        TEST_ID,
        {},
        dataset_fetcher=dataset_fetcher,
        test_data_downloader=test_data_downloader,
    )


def test_verify_hid_simple_fail():
    test_data_downloader = t_data_downloader_for(b"expected")
    dataset_fetcher = dataset_fetcher_for(TEST_ID, {None: b"unexpected"})
    assertion_error = None
    try:
        verify_hid(
            "",
            TEST_ID,
            {},
            dataset_fetcher=dataset_fetcher,
            test_data_downloader=test_data_downloader,
        )
    except AssertionError as ae:
        assertion_error = ae
    assert assertion_error


def test_verify_extra_files_success(tmp_path):
    test_data_downloader = t_data_downloader_for(
        {
            "": b"expected",
            "test_extras/expected.txt": b"extra_expected",
        }
    )
    found_data = {
        None: b"expected",
        "extra_1": b"extra_expected",
    }
    extra_files = [{"type": "file", "name": "extra_1", "attributes": {}, "value": "test_extras/expected.txt"}]
    dataset_fetcher = dataset_fetcher_for(TEST_ID, found_data)
    assert not (tmp_path / "expected.txt").exists()
    verify_hid(
        "",
        TEST_ID,
        {"extra_files": extra_files},
        dataset_fetcher=dataset_fetcher,
        test_data_downloader=test_data_downloader,
        keep_outputs_dir=tmp_path,
    )
    extra_path = tmp_path / "test_extras" / "expected.txt"
    assert extra_path.exists()
    assert extra_path.open("rb").read() == b"extra_expected"


def test_verify_extra_files_failed(tmp_path):
    test_data_downloader = t_data_downloader_for(
        {
            "": b"expected",
            "test_extras/expected.txt": b"extra_expected",
        }
    )
    found_data = {
        None: b"expected",
        "extra_1": b"extra_unexpected",
    }
    extra_files = [{"type": "file", "name": "extra_1", "attributes": {}, "value": "test_extras/expected.txt"}]
    dataset_fetcher = dataset_fetcher_for(TEST_ID, found_data)
    assert not (tmp_path / "expected.txt").exists()
    assert_error = None
    try:
        verify_hid(
            "",
            TEST_ID,
            {"extra_files": extra_files},
            dataset_fetcher=dataset_fetcher,
            test_data_downloader=test_data_downloader,
            keep_outputs_dir=tmp_path,
        )
    except AssertionError as ae:
        assert_error = ae
    assert assert_error
    assert "Composite file (extra_1) of History item mydatasetid different than expected" in str(assert_error)
    extra_path = tmp_path / "test_extras" / "expected.txt"
    assert extra_path.exists()
    assert extra_path.open("rb").read() == b"extra_unexpected"



from galaxy.tool_util.deps.singularity_util import build_singularity_run_command


def test_build_singularity_run_command_defaults():
    cmd = build_singularity_run_command(
        container_command="echo hi",
        image="busybox",
    )
    assert cmd == "singularity -s exec --contain --cleanenv --ipc --pid --no-mount tmp busybox echo hi"


def test_build_singularity_run_command_no_cleanenv_ipc_pid():
    cmd = build_singularity_run_command(
        container_command="echo hi",
        image="busybox",
        cleanenv=False,
        pid=False,
        ipc=False,
    )
    assert cmd == "singularity -s exec --contain --no-mount tmp busybox echo hi"



# Not yet tested:
# - 21_09_fix_from_work_dir_whitespace
# - 23_0_consider_optional_text
import os
from typing import List

import pytest

from galaxy.tool_util.unittest_utils import functional_test_tool_path
from galaxy.tool_util.upgrade import (
    Advice,
    advise_on_upgrade,
    TARGET_TOO_NEW,
)

FUTURE_GALAXY_VERSION = "29.6"


def test_does_not_work_on_non_xml_tools():
    simple_constructs = _tool_path("simple_constructs.yml")
    with pytest.raises(Exception) as e:
        advise_on_upgrade(simple_constructs, "16.04")
    assert "XML tools" in str(e)


def test_does_not_work_on_versions_too_new():
    simple_constructs = _tool_path("simple_constructs.xml")
    with pytest.raises(Exception) as e:
        advise_on_upgrade(simple_constructs, FUTURE_GALAXY_VERSION)
    assert TARGET_TOO_NEW in str(e)


def test_old_advice_does_not_appear_when_upgrading_past_it():
    a_16_01_tool = _tool_path("tool_provided_metadata_1.xml")
    advice = advise_on_upgrade(a_16_01_tool)
    assert_has_advice(advice, "16_04_consider_implicit_extra_file_collection")

    a_17_09_tool = _tool_path("tool_provided_metadata_6.xml")
    advice = advise_on_upgrade(a_17_09_tool)
    assert_not_has_advice(advice, "16_04_consider_implicit_extra_file_collection")


def test_interpreter_advice_positive():
    legacy_interpreter = _tool_path("legacy_interpreter.xml")
    advice = advise_on_upgrade(legacy_interpreter, "16.04")
    assert_has_advice(advice, "16_04_fix_interpreter")
    assert_not_has_advice(advice, "16_04_ready_interpreter")


def test_interpreter_advice_negative():
    simple_constructs = _tool_path("simple_constructs.xml")
    advice = advise_on_upgrade(simple_constructs, "16.04")
    assert_not_has_advice(advice, "16_04_fix_interpreter")
    assert_has_advice(advice, "16_04_ready_interpreter")


def test_output_format_advice_positive():
    output_format = _tool_path("output_format.xml")
    advice = advise_on_upgrade(output_format, "16.04")
    assert_has_advice(advice, "16_04_fix_output_format")


def test_20_05_inputs_changes():
    inputs_as_json = _tool_path("inputs_as_json.xml")
    advice = advise_on_upgrade(inputs_as_json)
    assert_has_advice(advice, "20_05_consider_inputs_as_json_changes")

    simple_constructs = _tool_path("simple_constructs.xml")
    advice = advise_on_upgrade(simple_constructs)
    assert_not_has_advice(advice, "20_05_consider_inputs_as_json_changes")


def test_output_format_advice_negative():
    simple_constructs = _tool_path("simple_constructs.xml")
    advice = advise_on_upgrade(simple_constructs, "16.04")
    assert_not_has_advice(advice, "16_04_fix_output_format")


def test_consider_implicit_extra_file_collection():
    simple_constructs = _tool_path("simple_constructs.xml")
    advice = advise_on_upgrade(simple_constructs, "16.04")
    assert_has_advice(advice, "16_04_consider_implicit_extra_file_collection")


def test_consider_stdio():
    simple_constructs = _tool_path("simple_constructs.xml")
    advice = advise_on_upgrade(simple_constructs, "16.04")
    assert_has_advice(advice, "16_04_exit_code")

    a_17_09_tool = _tool_path("tool_provided_metadata_6.xml")
    advice = advise_on_upgrade(a_17_09_tool)
    assert_not_has_advice(advice, "16_04_exit_code")


def test_2009_output_collection_advice_positive():
    collection_creates_list = _tool_path("collection_creates_list.xml")
    advice = advise_on_upgrade(collection_creates_list)
    assert_has_advice(advice, "20_09_consider_output_collection_order")


def test_2009_output_collection_advice_negative():
    simple_constructs = _tool_path("simple_constructs.xml")
    advice = advise_on_upgrade(simple_constructs)
    assert_not_has_advice(advice, "20_09_consider_output_collection_order")


def test_2009_consider_strict_shell_positive():
    simple_constructs = _tool_path("simple_constructs.xml")
    advice = advise_on_upgrade(simple_constructs)
    assert_has_advice(advice, "20_09_consider_set_e")


def test_1801_consider_home_directory():
    simple_constructs = _tool_path("simple_constructs.xml")
    advice = advise_on_upgrade(simple_constructs)
    assert_has_advice(advice, "18_01_consider_home_directory")


def test_1801_consider_structured_like():
    simple_constructs = _tool_path("collection_paired_conditional_structured_like.xml")
    advice = advise_on_upgrade(simple_constructs)
    assert_has_advice(advice, "18_01_consider_structured_like")

    simple_constructs = _tool_path("simple_constructs.xml")
    advice = advise_on_upgrade(simple_constructs)
    assert_not_has_advice(advice, "18_01_consider_structured_like")


def test_21_09_data_source_advice():
    test_data_source = _tool_path("test_data_source.xml")
    advice = advise_on_upgrade(test_data_source)
    assert_has_advice(advice, "21_09_consider_python_environment")

    simple_constructs = _tool_path("simple_constructs.xml")
    advice = advise_on_upgrade(simple_constructs)
    assert_not_has_advice(advice, "21_09_consider_python_environment")


def test_24_0_request_cleaning():
    test_data_source = _tool_path("test_data_source.xml")
    advice = advise_on_upgrade(test_data_source)
    assert_has_advice(advice, "24_0_request_cleaning")

    simple_constructs = _tool_path("simple_constructs.xml")
    advice = advise_on_upgrade(simple_constructs)
    assert_not_has_advice(advice, "24_0_request_cleaning")


def test_24_2_test_case_validation():
    test_data_source = _tool_path("column_param.xml")
    advice = advise_on_upgrade(test_data_source)
    assert_has_advice(advice, "24_2_fix_test_case_validation")

    int_param = _tool_path("parameters/gx_int.xml")
    advice = advise_on_upgrade(int_param)
    assert_not_has_advice(advice, "24_2_fix_test_case_validation")


def _tool_path(tool_name: str):
    return os.path.join(functional_test_tool_path(tool_name))


def assert_has_advice(advice_list: List[Advice], advice_code: str):
    for advice in advice_list:
        if advice.advice_code["name"] == advice_code:
            return

    raise AssertionError(f"Was expecting advice {advice_code} in list of upgrade advice {advice_list}")


def assert_not_has_advice(advice_list: List[Advice], advice_code: str):
    for advice in advice_list:
        if advice.advice_code["name"] == advice_code:
            raise AssertionError(f"Was not expecting advice {advice_code} in list of upgrade advice {advice_list}")



import pytest

from galaxy.tool_util.deps.requirements import ContainerDescription


@pytest.mark.parametrize(
    "identifier,expected_identifier",
    [
        ("mulled-abc", "mulled-abc"),
        ("docker://mulled-abc", "docker://mulled-abc"),
        ("/Cache/mulled-abc", "/Cache/mulled-abc"),
        ("/Cache/mUlled-abc", "/Cache/mulled-abc"),
        ("", ""),
    ],
)
def test_container_description(identifier, expected_identifier):
    assert ContainerDescription(identifier=identifier).identifier == expected_identifier


def test_to_from_dict():
    container_description_dict = ContainerDescription("mulled-abc").to_dict()
    container_description = ContainerDescription.from_dict(container_description_dict)
    assert container_description_dict == container_description.to_dict()



from typing import (
    Any,
    Dict,
    Optional,
)

from galaxy.tool_util.parameters import (
    DataRequestCollectionUri,
    DataRequestInternalHda,
    DataRequestInternalHdca,
    DataRequestUri,
    decode,
    dereference,
    encode,
    fill_static_defaults,
    input_models_for_tool_source,
    landing_decode,
    landing_encode,
    LandingRequestToolState,
    RelaxedRequestToolState,
    RequestInternalDereferencedToolState,
    RequestInternalToolState,
    RequestToolState,
    strictify,
)
from galaxy.tool_util.parser.util import parse_profile_version
from .test_parameter_test_cases import tool_source_for

EXAMPLE_ID_1_ENCODED = "123456789abcde"
EXAMPLE_ID_1 = 13
EXAMPLE_ID_2_ENCODED = "123456789abcd2"
EXAMPLE_ID_2 = 14

ID_MAP: Dict[int, str] = {
    EXAMPLE_ID_1: EXAMPLE_ID_1_ENCODED,
    EXAMPLE_ID_2: EXAMPLE_ID_2_ENCODED,
}


def test_decode_data():
    tool_source = tool_source_for("parameters/gx_data")
    bundle = input_models_for_tool_source(tool_source)
    request_state = RequestToolState({"parameter": {"src": "hda", "id": EXAMPLE_ID_1_ENCODED}})
    request_state.validate(bundle)
    decoded_state = decode(request_state, bundle, _fake_decode)
    assert decoded_state.input_state["parameter"]["src"] == "hda"
    assert decoded_state.input_state["parameter"]["id"] == EXAMPLE_ID_1


def test_decode_data_batch():
    tool_source = tool_source_for("parameters/gx_data")
    bundle = input_models_for_tool_source(tool_source)
    request_state = RequestToolState(
        {"parameter": {"__class__": "Batch", "values": [{"src": "hda", "id": EXAMPLE_ID_1_ENCODED}]}}
    )
    request_state.validate(bundle)
    decoded_state = decode(request_state, bundle, _fake_decode)
    assert decoded_state.input_state["parameter"]["values"][0]["src"] == "hda"
    assert decoded_state.input_state["parameter"]["values"][0]["id"] == EXAMPLE_ID_1


def test_decode_collection():
    tool_source = tool_source_for("parameters/gx_data_collection")
    bundle = input_models_for_tool_source(tool_source)
    request_state = RequestToolState({"parameter": {"src": "hdca", "id": EXAMPLE_ID_1_ENCODED}})
    request_state.validate(bundle)
    decoded_state = decode(request_state, bundle, _fake_decode)
    assert decoded_state.input_state["parameter"]["src"] == "hdca"
    assert decoded_state.input_state["parameter"]["id"] == EXAMPLE_ID_1


def test_decode_repeat():
    tool_source = tool_source_for("parameters/gx_repeat_data")
    bundle = input_models_for_tool_source(tool_source)
    request_state = RequestToolState({"parameter": [{"data_parameter": {"src": "hda", "id": EXAMPLE_ID_1_ENCODED}}]})
    request_state.validate(bundle)
    decoded_state = decode(request_state, bundle, _fake_decode)
    assert decoded_state.input_state["parameter"][0]["data_parameter"]["src"] == "hda"
    assert decoded_state.input_state["parameter"][0]["data_parameter"]["id"] == EXAMPLE_ID_1


def test_decode_section():
    tool_source = tool_source_for("parameters/gx_section_data")
    bundle = input_models_for_tool_source(tool_source)
    request_state = RequestToolState({"parameter": {"data_parameter": {"src": "hda", "id": EXAMPLE_ID_1_ENCODED}}})
    request_state.validate(bundle)
    decoded_state = decode(request_state, bundle, _fake_decode)
    assert decoded_state.input_state["parameter"]["data_parameter"]["src"] == "hda"
    assert decoded_state.input_state["parameter"]["data_parameter"]["id"] == EXAMPLE_ID_1


def test_decode_conditional():
    tool_source = tool_source_for("identifier_in_conditional")
    bundle = input_models_for_tool_source(tool_source)
    request_state = RequestToolState(
        {"outer_cond": {"multi_input": False, "input1": {"src": "hda", "id": EXAMPLE_ID_1_ENCODED}}}
    )
    request_state.validate(bundle)
    decoded_state = decode(request_state, bundle, _fake_decode)
    assert decoded_state.input_state["outer_cond"]["input1"]["src"] == "hda"
    assert decoded_state.input_state["outer_cond"]["input1"]["id"] == EXAMPLE_ID_1


def test_multi_data():
    tool_source = tool_source_for("parameters/gx_data_multiple")
    bundle = input_models_for_tool_source(tool_source)
    request_state = RequestToolState(
        {"parameter": [{"src": "hda", "id": EXAMPLE_ID_1_ENCODED}, {"src": "hda", "id": EXAMPLE_ID_2_ENCODED}]}
    )
    request_state.validate(bundle)
    decoded_state = decode(request_state, bundle, _fake_decode)
    assert decoded_state.input_state["parameter"][0]["src"] == "hda"
    assert decoded_state.input_state["parameter"][0]["id"] == EXAMPLE_ID_1
    assert decoded_state.input_state["parameter"][1]["src"] == "hda"
    assert decoded_state.input_state["parameter"][1]["id"] == EXAMPLE_ID_2

    encoded_state = encode(decoded_state, bundle, _fake_encode)
    assert encoded_state.input_state["parameter"][0]["src"] == "hda"
    assert encoded_state.input_state["parameter"][0]["id"] == EXAMPLE_ID_1_ENCODED
    assert encoded_state.input_state["parameter"][1]["src"] == "hda"
    assert encoded_state.input_state["parameter"][1]["id"] == EXAMPLE_ID_2_ENCODED


def test_landing_encode_data():
    tool_source = tool_source_for("parameters/gx_data")
    bundle = input_models_for_tool_source(tool_source)
    request_state = LandingRequestToolState({"parameter": {"src": "hda", "id": EXAMPLE_ID_1_ENCODED}})
    request_state.validate(bundle)
    decoded_state = landing_decode(request_state, bundle, _fake_decode)
    assert decoded_state.input_state["parameter"]["src"] == "hda"
    assert decoded_state.input_state["parameter"]["id"] == EXAMPLE_ID_1

    encoded_state = landing_encode(decoded_state, bundle, _fake_encode)
    assert encoded_state.input_state["parameter"]["src"] == "hda"
    assert encoded_state.input_state["parameter"]["id"] == EXAMPLE_ID_1_ENCODED


def test_landing_encode_data_batch():
    tool_source = tool_source_for("parameters/gx_data")
    bundle = input_models_for_tool_source(tool_source)
    request_state = LandingRequestToolState(
        {"parameter": {"__class__": "Batch", "values": [{"src": "hda", "id": EXAMPLE_ID_1_ENCODED}]}}
    )
    request_state.validate(bundle)
    decoded_state = landing_decode(request_state, bundle, _fake_decode)
    assert decoded_state.input_state["parameter"]["values"][0]["src"] == "hda"
    assert decoded_state.input_state["parameter"]["values"][0]["id"] == EXAMPLE_ID_1

    encoded_state = landing_encode(decoded_state, bundle, _fake_encode)
    assert encoded_state.input_state["parameter"]["values"][0]["src"] == "hda"
    assert encoded_state.input_state["parameter"]["values"][0]["id"] == EXAMPLE_ID_1_ENCODED


def test_dereference():
    tool_source = tool_source_for("parameters/gx_data")
    bundle = input_models_for_tool_source(tool_source)
    raw_request_state = {"parameter": {"src": "url", "url": "gxfiles://mystorage/1.bed", "ext": "bed"}}
    request_state = RequestInternalToolState(raw_request_state)
    request_state.validate(bundle)

    exception: Optional[Exception] = None
    try:
        # quickly verify this request needs to be dereferenced
        bad_state = RequestInternalDereferencedToolState(raw_request_state)
        bad_state.validate(bundle)
    except Exception as e:
        exception = e
    assert exception is not None

    dereferenced_state = dereference(request_state, bundle, _fake_dereference, _fake_collection_deference)
    assert isinstance(dereferenced_state, RequestInternalDereferencedToolState)
    dereferenced_state.validate(bundle)


def test_fill_defaults():
    with_defaults = fill_state_for({}, "parameters/gx_int")
    assert with_defaults["parameter"] == 1
    with_defaults = fill_state_for({}, "parameters/gx_float")
    assert with_defaults["parameter"] == 1.0
    with_defaults = fill_state_for({}, "parameters/gx_boolean")
    assert with_defaults["parameter"] is False
    with_defaults = fill_state_for({}, "parameters/gx_boolean_optional")
    # This is False unfortunately - see comments in gx_boolean_optional XML.
    assert with_defaults["parameter"] is False
    with_defaults = fill_state_for({}, "parameters/gx_boolean_checked")
    assert with_defaults["parameter"] is True
    with_defaults = fill_state_for({}, "parameters/gx_boolean_optional_checked")
    assert with_defaults["parameter"] is True

    with_defaults = fill_state_for({}, "parameters/gx_conditional_boolean")
    assert with_defaults["conditional_parameter"]["test_parameter"] is False
    assert with_defaults["conditional_parameter"]["boolean_parameter"] is False

    with_defaults = fill_state_for({"conditional_parameter": {}}, "parameters/gx_conditional_boolean")
    assert with_defaults["conditional_parameter"]["test_parameter"] is False
    assert with_defaults["conditional_parameter"]["boolean_parameter"] is False

    with_defaults = fill_state_for({}, "parameters/gx_repeat_boolean")
    assert len(with_defaults["parameter"]) == 0
    with_defaults = fill_state_for({"parameter": [{}]}, "parameters/gx_repeat_boolean")
    assert len(with_defaults["parameter"]) == 1
    instance_state = with_defaults["parameter"][0]
    assert instance_state["boolean_parameter"] is False

    with_defaults = fill_state_for({}, "parameters/gx_repeat_boolean_min")
    assert len(with_defaults["parameter"]) == 2
    assert with_defaults["parameter"][0]["boolean_parameter"] is False
    assert with_defaults["parameter"][1]["boolean_parameter"] is False
    with_defaults = fill_state_for({"parameter": [{}, {}]}, "parameters/gx_repeat_boolean_min")
    assert len(with_defaults["parameter"]) == 2
    assert with_defaults["parameter"][0]["boolean_parameter"] is False
    assert with_defaults["parameter"][1]["boolean_parameter"] is False
    with_defaults = fill_state_for({"parameter": [{}]}, "parameters/gx_repeat_boolean_min")
    assert with_defaults["parameter"][0]["boolean_parameter"] is False
    assert with_defaults["parameter"][1]["boolean_parameter"] is False

    with_defaults = fill_state_for({}, "parameters/gx_section_boolean")
    assert with_defaults["parameter"]["boolean_parameter"] is False

    with_defaults = fill_state_for({}, "parameters/gx_drill_down_exact_with_selection")
    assert with_defaults["parameter"] == "aba"

    with_defaults = fill_state_for({}, "parameters/gx_hidden")
    assert with_defaults["parameter"] == "moo"

    with_defaults = fill_state_for({}, "parameters/gx_genomebuild_optional")
    assert with_defaults["parameter"] is None

    with_defaults = fill_state_for({}, "parameters/gx_select")
    assert with_defaults["parameter"] == "--ex1"

    with_defaults = fill_state_for({}, "parameters/gx_select_optional")
    assert with_defaults["parameter"] is None

    with_defaults = fill_state_for({}, "parameters/gx_select_multiple")
    assert with_defaults["parameter"] == []

    with_defaults = fill_state_for({}, "parameters/gx_select_multiple_optional")
    assert with_defaults["parameter"] == []

    with_defaults = fill_state_for({}, "parameters/gx_select_multiple_one_default")
    assert with_defaults["parameter"] == ["--ex3"]

    # Do not fill in dynamic defaults... these require a Galaxy runtime.
    with_defaults = fill_state_for({}, "remove_value", partial=True)
    assert "choose_value" not in with_defaults

    with_defaults = fill_state_for(
        {"single": {"src": "hda", "id": 4}}, "select_from_dataset_in_conditional", partial=True
    )
    assert with_defaults["cond"]["cond"] == "single"
    assert with_defaults["cond"]["inner_cond"]["inner_cond"] == "single"

    # dynamic parameters should just stay empty - null would cause runtime to skip over population
    with_defaults = fill_state_for({}, "parameters/gx_select_dynamic", partial=True)
    assert "parameter" not in with_defaults

    # dynamic parameters should just stay empty - null would cause runtime to skip over population
    with_defaults = fill_state_for(
        {"conditional_parameter": {"test_parameter": False}},
        "parameters/gx_conditional_boolean_discriminate_on_string_value",
    )
    assert "conditional_parameter" in with_defaults
    assert "boolean_parameter" in with_defaults["conditional_parameter"]
    assert with_defaults["conditional_parameter"]["boolean_parameter"] is False


def test_strictify():
    strict_state = strictify_for({"parameter": 1}, "parameters/gx_int")
    assert strict_state["parameter"] == 1

    strict_state = strictify_for({}, "parameters/gx_text_optional_false")
    assert strict_state["parameter"] == ""

    strict_state = strictify_for({"parameter": None}, "parameters/gx_text_optional_false")
    assert strict_state["parameter"] == ""


def strictify_for(tool_state: Dict[str, Any], tool_path: str) -> Dict[str, Any]:
    tool_source = tool_source_for(tool_path)
    bundle = input_models_for_tool_source(tool_source)
    relaxed_state = RelaxedRequestToolState(tool_state)
    relaxed_state.validate(bundle)
    return strictify(relaxed_state, bundle).input_state


def _fake_dereference(input: DataRequestUri) -> DataRequestInternalHda:
    return DataRequestInternalHda(id=EXAMPLE_ID_1, src="hda")


def _fake_collection_deference(input: DataRequestCollectionUri) -> DataRequestInternalHdca:
    return DataRequestInternalHdca(id=EXAMPLE_ID_1, src="hdca")


def _fake_decode(input: str) -> int:
    return next(key for key, value in ID_MAP.items() if value == input)


def _fake_encode(input: int) -> str:
    return ID_MAP[input]


def fill_state_for(tool_state: Dict[str, Any], tool_path: str, partial: bool = False) -> Dict[str, Any]:
    tool_source = tool_source_for(tool_path)
    bundle = input_models_for_tool_source(tool_source)
    profile = parse_profile_version(tool_source)
    internal_state = fill_static_defaults(tool_state, bundle, profile, partial=partial)
    return internal_state



import tempfile

from galaxy.tool_util.loader_directory import is_a_yaml_with_class


def test_is_a_yaml_with_class():
    with tempfile.NamedTemporaryFile("w", suffix=".yaml") as tf:
        fname = tf.name
        tf.write("""class: GalaxyWorkflow
name: "Test Workflow"
inputs:
  - id: input1
outputs:
  - id: wf_output_1
    outputSource: first_cat/out_file1
steps:
  - tool_id: cat
    label: first_cat
    in:
      input1: input1""")
        tf.flush()
        assert is_a_yaml_with_class(fname, ["GalaxyWorkflow"])



import pydantic

from galaxy.tool_util.toolbox.views.definitions import (
    StaticToolBoxView,
    StaticToolBoxViewTypeEnum,
)

EXAMPLE_1 = {
    "id": "rna",
    "name": "RNA Seq",
    "type": "activity",
    "items": [
        {
            "type": "workflow",
            "id": "12345abcd",
        },
        {
            "type": "section",
            "name": "My Cool Section",
        },
        {
            "type": "label",
            "text": "My Cool Label",
            "id": "my-cool-label-id",
        },
        {
            "type": "tool",
            "id": "cat1",
        },
    ],
}


def test_root_parsing():
    view = StaticToolBoxView.from_dict(EXAMPLE_1)
    assert view.id == "rna"
    assert view.name == "RNA Seq"
    assert view.view_type == StaticToolBoxViewTypeEnum.activity
    assert len(view.items) == 4
    workflow_item = view.items[0]
    assert workflow_item.content_type == "workflow"
    assert workflow_item.id == "12345abcd"

    section_item = view.items[1]
    assert section_item.content_type == "section"
    assert section_item.id is None
    assert section_item.name == "My Cool Section"

    outer_label_item = view.items[2]
    assert outer_label_item.content_type == "label"
    assert outer_label_item.id == "my-cool-label-id"
    assert outer_label_item.text == "My Cool Label"

    output_tool_item = view.items[3]
    assert output_tool_item.content_type == "tool"
    assert output_tool_item.id == "cat1"


def test_section_parsing():
    view = StaticToolBoxView.from_dict(
        {
            "id": "rna",
            "name": "RNA Seq",
            "type": "activity",
            "items": [
                {
                    "type": "section",
                    "name": "My Cool Section",
                    "items": [
                        {
                            "type": "workflow",
                            "id": "123456",
                        }
                    ],
                },
            ],
        }
    )
    section_item = view.items[0]
    assert len(section_item.items) == 1
    assert section_item.items[0].content_type == "workflow"


def test_no_nested_sections():
    exception_raised = False
    try:
        StaticToolBoxView.from_dict(
            {
                "id": "rna",
                "name": "RNA Seq",
                "type": "activity",
                "items": [
                    {
                        "type": "section",
                        "name": "My Cool Section",
                        "items": [
                            {
                                "type": "section",
                                "name": "My Cool Inner Section",
                            },
                        ],
                    },
                ],
            }
        )
    except pydantic.ValidationError:
        exception_raised = True
    assert exception_raised
